// @generated by the capnpc-rust plugin to the Cap'n Proto schema compiler.
// DO NOT EDIT.
// source: src/keystore_protocol.capnp
// capnp binary version: 1.5.0
// capnpc crate version: 0.25.3



pub mod keystore {
    #![allow(unused_variables)]
    pub type BackendsParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::backends_params::Owned>;
    pub type BackendsResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::backends_results::Owned>;
    pub type DecryptParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::decrypt_params::Owned>;
    pub type DecryptResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::decrypt_results::Owned>;

    pub struct Client {
        pub client: ::capnp::capability::Client,
    }
    impl  ::capnp::capability::FromClientHook for Client {
        fn new(hook: Box<::capnp::capability::DynClientHook>) -> Self {
            Self { client: ::capnp::capability::Client::new(hook),  }
        }
        fn into_client_hook(self) -> Box<::capnp::capability::DynClientHook> {
            self.client.hook
        }
        fn as_client_hook(&self) -> &::capnp::capability::DynClientHook {
            &*self.client.hook
        }
    }
    #[derive(Copy, Clone)]
    pub struct Owned(());
    impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Capability.into() } }
    impl ::capnp::traits::Owned for Owned { type Reader<'a> = Client; type Builder<'a> = Client; }
    impl ::capnp::traits::Pipelined for Owned { type Pipeline = Client; }
    impl <'a,> ::capnp::traits::FromPointerReader<'a> for Client<>  {
        fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, _default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
            ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(reader.get_capability()?))
        }
    }
    impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Client<>  {
        fn init_pointer(_builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
            unimplemented!()
        }
        fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
            ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(builder.get_capability()?))
        }
    }

    impl <> ::capnp::traits::SetterInput<Owned<>> for Client<>  {
        fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, from: Self, _canonicalize: bool) -> ::capnp::Result<()> {
            pointer.set_capability(from.client.hook);
            ::core::result::Result::Ok(())
        }
    }
    impl  ::capnp::traits::HasTypeId for Client {
        const TYPE_ID: u64 = _private::TYPE_ID;
    }
    impl  Clone for Client {
        fn clone(&self) -> Self {
            Self { client: self.client.clone(),  }
        }
    }
    impl  Client {
        pub fn backends_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::backends_params::Owned,crate::keystore_protocol_capnp::keystore::backends_results::Owned> {
            self.client.new_call(_private::TYPE_ID, 0, ::core::option::Option::None)
        }
        pub fn decrypt_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::decrypt_params::Owned,crate::keystore_protocol_capnp::keystore::decrypt_results::Owned> {
            self.client.new_call(_private::TYPE_ID, 1, ::core::option::Option::None)
        }
    }
    pub trait Server<> : 'static  {
        fn backends(self: ::capnp::capability::Rc<Self>, _: BackendsParams<>, _: BackendsResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method keystore::Server::backends not implemented".to_string()))) }
        fn decrypt(self: ::capnp::capability::Rc<Self>, _: DecryptParams<>, _: DecryptResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method keystore::Server::decrypt not implemented".to_string()))) }
    }
    pub struct ServerDispatch<_T,> {
        pub server: ::capnp::capability::Rc<_T>,
    }
    impl <_S: Server + 'static, > ::capnp::capability::FromServer<_S> for Client   {
        type Dispatch = ServerDispatch<_S, >;
        fn from_server(s: ::capnp::capability::Rc<_S>) -> ServerDispatch<_S, > {
            ServerDispatch { server: s,  }
        }
    }
    impl <_T: Server> ::core::ops::Deref for ServerDispatch<_T> {
        type Target = _T;
        fn deref(&self) -> &_T { &self.server}
    }
    impl <_T: Server> ::core::clone::Clone for ServerDispatch<_T> {
        fn clone(&self) -> Self { Self { server: self.server.clone(),  } }
    }
    impl <_T: Server> ::capnp::capability::Server for ServerDispatch<_T> {
        fn dispatch_call(self, interface_id: u64, method_id: u16, params: ::capnp::capability::Params<::capnp::any_pointer::Owned>, results: ::capnp::capability::Results<::capnp::any_pointer::Owned>) -> ::capnp::capability::DispatchCallResult {
            match interface_id {
                _private::TYPE_ID => Self::dispatch_call_internal(self.server, method_id, params, results),
                _ => { ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::err(::capnp::Error::unimplemented("Method not implemented.".to_string())), false) }
            }
        }
        fn as_ptr(&self) -> usize { ::capnp::capability::Rc::as_ptr(&self.server) as usize }
    }
    impl <_T :Server> ServerDispatch<_T> {
        pub fn dispatch_call_internal(this: ::capnp::capability::Rc<_T>, method_id: u16, params: ::capnp::capability::Params<::capnp::any_pointer::Owned>, results: ::capnp::capability::Results<::capnp::any_pointer::Owned>) -> ::capnp::capability::DispatchCallResult {
            match method_id {
                0 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::backends(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                1 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::decrypt(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                _ => { ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::err(::capnp::Error::unimplemented("Method not implemented.".to_string())), false) }
            }
        }
    }
    pub(crate) mod _private {
        pub(crate) const TYPE_ID: u64 = 0xd7f0_750f_4788_65af;
    }

    pub mod inaccessible_decryption_key {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn get_key_descriptor(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::key_descriptor::Reader<'a>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_key_descriptor(&self) -> bool {
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn get_pkesk(self) -> ::capnp::Result<::capnp::data::Reader<'a>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(1), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_pkesk(&self) -> bool {
                !self.reader.get_pointer_field(1).is_null()
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 2 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn get_key_descriptor(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::key_descriptor::Builder<'a>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_key_descriptor(&mut self, value: crate::keystore_protocol_capnp::keystore::key_descriptor::Reader<'_>) -> ::capnp::Result<()> {
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_key_descriptor(self, ) -> crate::keystore_protocol_capnp::keystore::key_descriptor::Builder<'a> {
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
            }
            #[inline]
            pub fn has_key_descriptor(&self) -> bool {
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn get_pkesk(self) -> ::capnp::Result<::capnp::data::Builder<'a>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(1), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_pkesk(&mut self, value: ::capnp::data::Reader<'_>)  {
                self.builder.reborrow().get_pointer_field(1).set_data(value);
            }
            #[inline]
            pub fn init_pkesk(self, size: u32) -> ::capnp::data::Builder<'a> {
                self.builder.get_pointer_field(1).init_data(size)
            }
            #[inline]
            pub fn has_pkesk(&self) -> bool {
                !self.builder.is_pointer_field_null(1)
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
            pub fn get_key_descriptor(&self) -> crate::keystore_protocol_capnp::keystore::key_descriptor::Pipeline {
                ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
            }
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 54] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(183, 143, 143, 82, 232, 118, 133, 161),
                ::capnp::word(37, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(2, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(44, 3, 0, 0, 4, 4, 0, 0),
                ::capnp::word(21, 0, 0, 0, 250, 1, 0, 0),
                ::capnp::word(49, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(45, 0, 0, 0, 119, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 73, 110, 97),
                ::capnp::word(99, 99, 101, 115, 115, 105, 98, 108),
                ::capnp::word(101, 68, 101, 99, 114, 121, 112, 116),
                ::capnp::word(105, 111, 110, 75, 101, 121, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 114, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(40, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(52, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(49, 0, 0, 0, 50, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(44, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(56, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(107, 101, 121, 68, 101, 115, 99, 114),
                ::capnp::word(105, 112, 116, 111, 114, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(142, 90, 201, 49, 99, 219, 111, 141),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(112, 107, 101, 115, 107, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <crate::keystore_protocol_capnp::keystore::key_descriptor::Owned as ::capnp::introspect::Introspect>::introspect(),
                    1 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[0,1];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0,1];
            pub(crate) const TYPE_ID: u64 = 0xa185_76e8_528f_8fb7;
        }
    }

    pub mod decrypted_pkesk {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn get_index(self) -> u32 {
                self.reader.get_data_field::<u32>(0)
            }
            #[inline]
            pub fn get_fingerprint(self) -> ::capnp::Result<::capnp::data::Reader<'a>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_fingerprint(&self) -> bool {
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn get_algo(self) -> u8 {
                self.reader.get_data_field::<u8>(4)
            }
            #[inline]
            pub fn get_session_key(self) -> ::capnp::Result<::capnp::data::Reader<'a>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(1), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_session_key(&self) -> bool {
                !self.reader.get_pointer_field(1).is_null()
            }
            #[inline]
            pub fn get_fingerprint_version(self) -> u8 {
                self.reader.get_data_field::<u8>(5)
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 2 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn get_index(self) -> u32 {
                self.builder.get_data_field::<u32>(0)
            }
            #[inline]
            pub fn set_index(&mut self, value: u32)  {
                self.builder.set_data_field::<u32>(0, value);
            }
            #[inline]
            pub fn get_fingerprint(self) -> ::capnp::Result<::capnp::data::Builder<'a>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_fingerprint(&mut self, value: ::capnp::data::Reader<'_>)  {
                self.builder.reborrow().get_pointer_field(0).set_data(value);
            }
            #[inline]
            pub fn init_fingerprint(self, size: u32) -> ::capnp::data::Builder<'a> {
                self.builder.get_pointer_field(0).init_data(size)
            }
            #[inline]
            pub fn has_fingerprint(&self) -> bool {
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn get_algo(self) -> u8 {
                self.builder.get_data_field::<u8>(4)
            }
            #[inline]
            pub fn set_algo(&mut self, value: u8)  {
                self.builder.set_data_field::<u8>(4, value);
            }
            #[inline]
            pub fn get_session_key(self) -> ::capnp::Result<::capnp::data::Builder<'a>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(1), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_session_key(&mut self, value: ::capnp::data::Reader<'_>)  {
                self.builder.reborrow().get_pointer_field(1).set_data(value);
            }
            #[inline]
            pub fn init_session_key(self, size: u32) -> ::capnp::data::Builder<'a> {
                self.builder.get_pointer_field(1).init_data(size)
            }
            #[inline]
            pub fn has_session_key(&self) -> bool {
                !self.builder.is_pointer_field_null(1)
            }
            #[inline]
            pub fn get_fingerprint_version(self) -> u8 {
                self.builder.get_data_field::<u8>(5)
            }
            #[inline]
            pub fn set_fingerprint_version(&mut self, value: u8)  {
                self.builder.set_data_field::<u8>(5, value);
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 101] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(135, 169, 134, 148, 114, 56, 38, 210),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(2, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(8, 4, 0, 0, 75, 6, 0, 0),
                ::capnp::word(21, 0, 0, 0, 162, 1, 0, 0),
                ::capnp::word(45, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 31, 1, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 68, 101, 99),
                ::capnp::word(114, 121, 112, 116, 101, 100, 80, 107),
                ::capnp::word(101, 115, 107, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(20, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(125, 0, 0, 0, 50, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(120, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(132, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(129, 0, 0, 0, 98, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(128, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(140, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(3, 0, 0, 0, 4, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 2, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(137, 0, 0, 0, 42, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(132, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(144, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(4, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 3, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(141, 0, 0, 0, 90, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(140, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(152, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(2, 0, 0, 0, 5, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 4, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(149, 0, 0, 0, 154, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(152, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(164, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(105, 110, 100, 101, 120, 0, 0, 0),
                ::capnp::word(8, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(8, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(102, 105, 110, 103, 101, 114, 112, 114),
                ::capnp::word(105, 110, 116, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(97, 108, 103, 111, 0, 0, 0, 0),
                ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 101, 115, 115, 105, 111, 110, 75),
                ::capnp::word(101, 121, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(102, 105, 110, 103, 101, 114, 112, 114),
                ::capnp::word(105, 110, 116, 86, 101, 114, 115, 105),
                ::capnp::word(111, 110, 0, 0, 0, 0, 0, 0),
                ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <u32 as ::capnp::introspect::Introspect>::introspect(),
                    1 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                    2 => <u8 as ::capnp::introspect::Introspect>::introspect(),
                    3 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                    4 => <u8 as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[0,1,2,3,4];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[2,1,4,0,3];
            pub(crate) const TYPE_ID: u64 = 0xd226_3872_9486_a987;
        }
    }


    pub mod backend {
        #![allow(unused_variables)]
        pub type IdParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::backend::id_params::Owned>;
        pub type IdResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::backend::id_results::Owned>;
        pub type DevicesParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::backend::devices_params::Owned>;
        pub type DevicesResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::backend::devices_results::Owned>;
        pub type ImportParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::backend::import_params::Owned>;
        pub type ImportResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::backend::import_results::Owned>;

        pub struct Client {
            pub client: ::capnp::capability::Client,
        }
        impl  ::capnp::capability::FromClientHook for Client {
            fn new(hook: Box<::capnp::capability::DynClientHook>) -> Self {
                Self { client: ::capnp::capability::Client::new(hook),  }
            }
            fn into_client_hook(self) -> Box<::capnp::capability::DynClientHook> {
                self.client.hook
            }
            fn as_client_hook(&self) -> &::capnp::capability::DynClientHook {
                &*self.client.hook
            }
        }
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Capability.into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Client; type Builder<'a> = Client; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Client; }
        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Client<>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, _default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(reader.get_capability()?))
            }
        }
        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Client<>  {
            fn init_pointer(_builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                unimplemented!()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(builder.get_capability()?))
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Client<>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, from: Self, _canonicalize: bool) -> ::capnp::Result<()> {
                pointer.set_capability(from.client.hook);
                ::core::result::Result::Ok(())
            }
        }
        impl  ::capnp::traits::HasTypeId for Client {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl  Clone for Client {
            fn clone(&self) -> Self {
                Self { client: self.client.clone(),  }
            }
        }
        impl  Client {
            pub fn id_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::backend::id_params::Owned,crate::keystore_protocol_capnp::keystore::backend::id_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 0, ::core::option::Option::None)
            }
            pub fn devices_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::backend::devices_params::Owned,crate::keystore_protocol_capnp::keystore::backend::devices_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 1, ::core::option::Option::None)
            }
            pub fn import_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::backend::import_params::Owned,crate::keystore_protocol_capnp::keystore::backend::import_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 2, ::core::option::Option::None)
            }
        }
        pub trait Server<> : 'static  {
            fn id(self: ::capnp::capability::Rc<Self>, _: IdParams<>, _: IdResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method backend::Server::id not implemented".to_string()))) }
            fn devices(self: ::capnp::capability::Rc<Self>, _: DevicesParams<>, _: DevicesResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method backend::Server::devices not implemented".to_string()))) }
            fn import(self: ::capnp::capability::Rc<Self>, _: ImportParams<>, _: ImportResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method backend::Server::import not implemented".to_string()))) }
        }
        pub struct ServerDispatch<_T,> {
            pub server: ::capnp::capability::Rc<_T>,
        }
        impl <_S: Server + 'static, > ::capnp::capability::FromServer<_S> for Client   {
            type Dispatch = ServerDispatch<_S, >;
            fn from_server(s: ::capnp::capability::Rc<_S>) -> ServerDispatch<_S, > {
                ServerDispatch { server: s,  }
            }
        }
        impl <_T: Server> ::core::ops::Deref for ServerDispatch<_T> {
            type Target = _T;
            fn deref(&self) -> &_T { &self.server}
        }
        impl <_T: Server> ::core::clone::Clone for ServerDispatch<_T> {
            fn clone(&self) -> Self { Self { server: self.server.clone(),  } }
        }
        impl <_T: Server> ::capnp::capability::Server for ServerDispatch<_T> {
            fn dispatch_call(self, interface_id: u64, method_id: u16, params: ::capnp::capability::Params<::capnp::any_pointer::Owned>, results: ::capnp::capability::Results<::capnp::any_pointer::Owned>) -> ::capnp::capability::DispatchCallResult {
                match interface_id {
                    _private::TYPE_ID => Self::dispatch_call_internal(self.server, method_id, params, results),
                    _ => { ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::err(::capnp::Error::unimplemented("Method not implemented.".to_string())), false) }
                }
            }
            fn as_ptr(&self) -> usize { ::capnp::capability::Rc::as_ptr(&self.server) as usize }
        }
        impl <_T :Server> ServerDispatch<_T> {
            pub fn dispatch_call_internal(this: ::capnp::capability::Rc<_T>, method_id: u16, params: ::capnp::capability::Params<::capnp::any_pointer::Owned>, results: ::capnp::capability::Results<::capnp::any_pointer::Owned>) -> ::capnp::capability::DispatchCallResult {
                match method_id {
                    0 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::id(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    1 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::devices(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    2 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::import(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    _ => { ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::err(::capnp::Error::unimplemented("Method not implemented.".to_string())), false) }
                }
            }
        }
        pub(crate) mod _private {
            pub(crate) const TYPE_ID: u64 = 0x8bc8_412d_eace_e910;
        }

        pub mod id_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 20] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(251, 75, 9, 217, 123, 228, 255, 251),
                    ::capnp::word(45, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 186, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 66, 97, 99),
                    ::capnp::word(107, 101, 110, 100, 46, 105, 100, 36),
                    ::capnp::word(80, 97, 114, 97, 109, 115, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0xfbff_e47b_d909_4bfb;
            }
        }

        pub mod id_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,::capnp::text::Owned>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::text::Owned>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,::capnp::text::Owned>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::text::Owned> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<::capnp::text::Owned> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 48] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(248, 109, 127, 99, 103, 44, 245, 178),
                    ::capnp::word(45, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 194, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 66, 97, 99),
                    ::capnp::word(107, 101, 110, 100, 46, 105, 100, 36),
                    ::capnp::word(82, 101, 115, 117, 108, 116, 115, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(68, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<::capnp::text::Owned> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xb2f5_2c67_637f_6df8;
            }
        }

        pub mod devices_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 21] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(21, 133, 15, 43, 84, 233, 159, 150),
                    ::capnp::word(45, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 226, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 66, 97, 99),
                    ::capnp::word(107, 101, 110, 100, 46, 100, 101, 118),
                    ::capnp::word(105, 99, 101, 115, 36, 80, 97, 114),
                    ::capnp::word(97, 109, 115, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0x969f_e954_2b0f_8515;
            }
        }

        pub mod devices_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::device::Client>>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::device::Client>>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::device::Client>>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::device::Client>> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::device::Client>> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 53] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(195, 138, 115, 103, 130, 70, 211, 205),
                    ::capnp::word(45, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 234, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 66, 97, 99),
                    ::capnp::word(107, 101, 110, 100, 46, 100, 101, 118),
                    ::capnp::word(105, 99, 101, 115, 36, 82, 101, 115),
                    ::capnp::word(117, 108, 116, 115, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(84, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(14, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(17, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(83, 4, 206, 177, 4, 179, 139, 199),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::device::Client>> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xcdd3_4682_6773_8ac3;
            }
        }

        pub mod import_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_cert(self) -> ::capnp::Result<::capnp::data::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_cert(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_cert(self) -> ::capnp::Result<::capnp::data::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_cert(&mut self, value: ::capnp::data::Reader<'_>)  {
                    self.builder.reborrow().get_pointer_field(0).set_data(value);
                }
                #[inline]
                pub fn init_cert(self, size: u32) -> ::capnp::data::Builder<'a> {
                    self.builder.get_pointer_field(0).init_data(size)
                }
                #[inline]
                pub fn has_cert(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 37] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(227, 182, 58, 239, 62, 252, 6, 218),
                    ::capnp::word(45, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 218, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 66, 97, 99),
                    ::capnp::word(107, 101, 110, 100, 46, 105, 109, 112),
                    ::capnp::word(111, 114, 116, 36, 80, 97, 114, 97),
                    ::capnp::word(109, 115, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 42, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(20, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(99, 101, 114, 116, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xda06_fc3e_ef3a_b6e3;
            }
        }

        pub mod import_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::import_result::Owned>>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::import_result::Owned>>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::import_result::Owned>>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::import_result::Owned>> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::import_result::Owned>> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 53] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(200, 62, 98, 144, 30, 26, 30, 166),
                    ::capnp::word(45, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 226, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 66, 97, 99),
                    ::capnp::word(107, 101, 110, 100, 46, 105, 109, 112),
                    ::capnp::word(111, 114, 116, 36, 82, 101, 115, 117),
                    ::capnp::word(108, 116, 115, 0, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(84, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(14, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(95, 36, 3, 193, 63, 148, 84, 202),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::import_result::Owned>> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xa61e_1a1e_9062_3ec8;
            }
        }
    }

    pub mod import_result {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn get_key(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::key_descriptor::Reader<'a>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_key(&self) -> bool {
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn get_status(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::import_status::Reader<'a>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(1), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_status(&self) -> bool {
                !self.reader.get_pointer_field(1).is_null()
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 2 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn get_key(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::key_descriptor::Builder<'a>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_key(&mut self, value: crate::keystore_protocol_capnp::keystore::key_descriptor::Reader<'_>) -> ::capnp::Result<()> {
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_key(self, ) -> crate::keystore_protocol_capnp::keystore::key_descriptor::Builder<'a> {
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
            }
            #[inline]
            pub fn has_key(&self) -> bool {
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn get_status(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::import_status::Builder<'a>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(1), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_status(&mut self, value: crate::keystore_protocol_capnp::keystore::import_status::Reader<'_>) -> ::capnp::Result<()> {
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(1), value, false)
            }
            #[inline]
            pub fn init_status(self, ) -> crate::keystore_protocol_capnp::keystore::import_status::Builder<'a> {
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(1), 0)
            }
            #[inline]
            pub fn has_status(&self) -> bool {
                !self.builder.is_pointer_field_null(1)
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
            pub fn get_key(&self) -> crate::keystore_protocol_capnp::keystore::key_descriptor::Pipeline {
                ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
            }
            pub fn get_status(&self) -> crate::keystore_protocol_capnp::keystore::import_status::Pipeline {
                ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(1))
            }
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 52] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(95, 36, 3, 193, 63, 148, 84, 202),
                ::capnp::word(37, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(2, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(153, 10, 0, 0, 239, 10, 0, 0),
                ::capnp::word(21, 0, 0, 0, 146, 1, 0, 0),
                ::capnp::word(45, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 119, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 73, 109, 112),
                ::capnp::word(111, 114, 116, 82, 101, 115, 117, 108),
                ::capnp::word(116, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 34, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(36, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(48, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(45, 0, 0, 0, 58, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(40, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(52, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(107, 101, 121, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(142, 90, 201, 49, 99, 219, 111, 141),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 116, 97, 116, 117, 115, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(119, 192, 174, 250, 243, 251, 6, 137),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <crate::keystore_protocol_capnp::keystore::key_descriptor::Owned as ::capnp::introspect::Introspect>::introspect(),
                    1 => <crate::keystore_protocol_capnp::keystore::import_status::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[0,1];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0,1];
            pub(crate) const TYPE_ID: u64 = 0xca54_943f_c103_245f;
        }
    }

    pub mod import_status {
        pub use self::Which::{New,Updated,Unchanged};

        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <> Reader<'_,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichReader, ::capnp::NotInSchema> {
                match self.reader.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(New(
                            ()
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Updated(
                            ()
                        ))
                    }
                    2 => {
                        ::core::result::Result::Ok(Unchanged(
                            ()
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 0 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn set_new(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 0);
            }
            #[inline]
            pub fn set_updated(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 1);
            }
            #[inline]
            pub fn set_unchanged(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 2);
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichBuilder, ::capnp::NotInSchema> {
                match self.builder.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(New(
                            ()
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Updated(
                            ()
                        ))
                    }
                    2 => {
                        ::core::result::Result::Ok(Unchanged(
                            ()
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 68] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(119, 192, 174, 250, 243, 251, 6, 137),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(0, 0, 7, 0, 0, 0, 3, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(243, 10, 0, 0, 248, 11, 0, 0),
                ::capnp::word(21, 0, 0, 0, 146, 1, 0, 0),
                ::capnp::word(45, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 175, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 73, 109, 112),
                ::capnp::word(111, 114, 116, 83, 116, 97, 116, 117),
                ::capnp::word(115, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(12, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 255, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(69, 0, 0, 0, 34, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(64, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(76, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 254, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(73, 0, 0, 0, 66, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(68, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(80, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(2, 0, 253, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 2, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(77, 0, 0, 0, 82, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(76, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(88, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(110, 101, 119, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(117, 112, 100, 97, 116, 101, 100, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(117, 110, 99, 104, 97, 110, 103, 101),
                ::capnp::word(100, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <() as ::capnp::introspect::Introspect>::introspect(),
                    1 => <() as ::capnp::introspect::Introspect>::introspect(),
                    2 => <() as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[0,1,2];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0,2,1];
            pub(crate) const TYPE_ID: u64 = 0x8906_fbf3_faae_c077;
        }
        pub enum Which {
            New(()),
            Updated(()),
            Unchanged(()),
        }
        pub type WhichReader = Which;
        pub type WhichBuilder = Which;
    }


    pub mod device {
        #![allow(unused_variables)]
        pub type IdParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::device::id_params::Owned>;
        pub type IdResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::device::id_results::Owned>;
        pub type KeysParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::device::keys_params::Owned>;
        pub type KeysResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::device::keys_results::Owned>;

        pub struct Client {
            pub client: ::capnp::capability::Client,
        }
        impl  ::capnp::capability::FromClientHook for Client {
            fn new(hook: Box<::capnp::capability::DynClientHook>) -> Self {
                Self { client: ::capnp::capability::Client::new(hook),  }
            }
            fn into_client_hook(self) -> Box<::capnp::capability::DynClientHook> {
                self.client.hook
            }
            fn as_client_hook(&self) -> &::capnp::capability::DynClientHook {
                &*self.client.hook
            }
        }
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Capability.into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Client; type Builder<'a> = Client; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Client; }
        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Client<>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, _default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(reader.get_capability()?))
            }
        }
        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Client<>  {
            fn init_pointer(_builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                unimplemented!()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(builder.get_capability()?))
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Client<>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, from: Self, _canonicalize: bool) -> ::capnp::Result<()> {
                pointer.set_capability(from.client.hook);
                ::core::result::Result::Ok(())
            }
        }
        impl  ::capnp::traits::HasTypeId for Client {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl  Clone for Client {
            fn clone(&self) -> Self {
                Self { client: self.client.clone(),  }
            }
        }
        impl  Client {
            pub fn id_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::device::id_params::Owned,crate::keystore_protocol_capnp::keystore::device::id_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 0, ::core::option::Option::None)
            }
            pub fn keys_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::device::keys_params::Owned,crate::keystore_protocol_capnp::keystore::device::keys_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 1, ::core::option::Option::None)
            }
        }
        pub trait Server<> : 'static  {
            fn id(self: ::capnp::capability::Rc<Self>, _: IdParams<>, _: IdResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method device::Server::id not implemented".to_string()))) }
            fn keys(self: ::capnp::capability::Rc<Self>, _: KeysParams<>, _: KeysResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method device::Server::keys not implemented".to_string()))) }
        }
        pub struct ServerDispatch<_T,> {
            pub server: ::capnp::capability::Rc<_T>,
        }
        impl <_S: Server + 'static, > ::capnp::capability::FromServer<_S> for Client   {
            type Dispatch = ServerDispatch<_S, >;
            fn from_server(s: ::capnp::capability::Rc<_S>) -> ServerDispatch<_S, > {
                ServerDispatch { server: s,  }
            }
        }
        impl <_T: Server> ::core::ops::Deref for ServerDispatch<_T> {
            type Target = _T;
            fn deref(&self) -> &_T { &self.server}
        }
        impl <_T: Server> ::core::clone::Clone for ServerDispatch<_T> {
            fn clone(&self) -> Self { Self { server: self.server.clone(),  } }
        }
        impl <_T: Server> ::capnp::capability::Server for ServerDispatch<_T> {
            fn dispatch_call(self, interface_id: u64, method_id: u16, params: ::capnp::capability::Params<::capnp::any_pointer::Owned>, results: ::capnp::capability::Results<::capnp::any_pointer::Owned>) -> ::capnp::capability::DispatchCallResult {
                match interface_id {
                    _private::TYPE_ID => Self::dispatch_call_internal(self.server, method_id, params, results),
                    _ => { ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::err(::capnp::Error::unimplemented("Method not implemented.".to_string())), false) }
                }
            }
            fn as_ptr(&self) -> usize { ::capnp::capability::Rc::as_ptr(&self.server) as usize }
        }
        impl <_T :Server> ServerDispatch<_T> {
            pub fn dispatch_call_internal(this: ::capnp::capability::Rc<_T>, method_id: u16, params: ::capnp::capability::Params<::capnp::any_pointer::Owned>, results: ::capnp::capability::Results<::capnp::any_pointer::Owned>) -> ::capnp::capability::DispatchCallResult {
                match method_id {
                    0 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::id(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    1 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::keys(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    _ => { ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::err(::capnp::Error::unimplemented("Method not implemented.".to_string())), false) }
                }
            }
        }
        pub(crate) mod _private {
            pub(crate) const TYPE_ID: u64 = 0xc78b_b304_b1ce_0453;
        }

        pub mod id_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 20] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(176, 152, 255, 244, 187, 152, 168, 227),
                    ::capnp::word(44, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 178, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 68, 101, 118),
                    ::capnp::word(105, 99, 101, 46, 105, 100, 36, 80),
                    ::capnp::word(97, 114, 97, 109, 115, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0xe3a8_98bb_f4ff_98b0;
            }
        }

        pub mod id_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,::capnp::text::Owned>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::text::Owned>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,::capnp::text::Owned>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::text::Owned> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<::capnp::text::Owned> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 48] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(81, 38, 153, 31, 40, 95, 243, 131),
                    ::capnp::word(44, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 186, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 68, 101, 118),
                    ::capnp::word(105, 99, 101, 46, 105, 100, 36, 82),
                    ::capnp::word(101, 115, 117, 108, 116, 115, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(68, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<::capnp::text::Owned> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0x83f3_5f28_1f99_2651;
            }
        }

        pub mod keys_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 20] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(125, 95, 64, 253, 222, 11, 196, 217),
                    ::capnp::word(44, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 194, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 68, 101, 118),
                    ::capnp::word(105, 99, 101, 46, 107, 101, 121, 115),
                    ::capnp::word(36, 80, 97, 114, 97, 109, 115, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0xd9c4_0bde_fd40_5f7d;
            }
        }

        pub mod keys_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::key_descriptor::Owned>>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::key_descriptor::Owned>>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::key_descriptor::Owned>>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::key_descriptor::Owned>> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::key_descriptor::Owned>> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 53] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(233, 168, 46, 45, 21, 237, 235, 192),
                    ::capnp::word(44, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 202, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 68, 101, 118),
                    ::capnp::word(105, 99, 101, 46, 107, 101, 121, 115),
                    ::capnp::word(36, 82, 101, 115, 117, 108, 116, 115),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(84, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(14, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(142, 90, 201, 49, 99, 219, 111, 141),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::key_descriptor::Owned>> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xc0eb_ed15_2d2e_a8e9;
            }
        }
    }

    pub mod key_descriptor {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn get_handle(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::key::Client> {
                match self.reader.get_pointer_field(0).get_capability() { ::core::result::Result::Ok(c) => ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(c)), ::core::result::Result::Err(e) => ::core::result::Result::Err(e)}
            }
            #[inline]
            pub fn has_handle(&self) -> bool {
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn get_public_key(self) -> ::capnp::Result<::capnp::data::Reader<'a>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(1), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_public_key(&self) -> bool {
                !self.reader.get_pointer_field(1).is_null()
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 2 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn get_handle(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::key::Client> {
                match self.builder.get_pointer_field(0).get_capability() { ::core::result::Result::Ok(c) => ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(c)), ::core::result::Result::Err(e) => ::core::result::Result::Err(e)}
            }
            #[inline]
            pub fn set_handle(&mut self, value: crate::keystore_protocol_capnp::keystore::key::Client)  {
                self.builder.reborrow().get_pointer_field(0).set_capability(value.client.hook);
            }
            #[inline]
            pub fn has_handle(&self) -> bool {
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn get_public_key(self) -> ::capnp::Result<::capnp::data::Builder<'a>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(1), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_public_key(&mut self, value: ::capnp::data::Reader<'_>)  {
                self.builder.reborrow().get_pointer_field(1).set_data(value);
            }
            #[inline]
            pub fn init_public_key(self, size: u32) -> ::capnp::data::Builder<'a> {
                self.builder.get_pointer_field(1).init_data(size)
            }
            #[inline]
            pub fn has_public_key(&self) -> bool {
                !self.builder.is_pointer_field_null(1)
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
            pub fn get_handle(&self) -> crate::keystore_protocol_capnp::keystore::key::Client {
                ::capnp::capability::FromClientHook::new(self._typeless.get_pointer_field(0).as_cap())
            }
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 53] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(142, 90, 201, 49, 99, 219, 111, 141),
                ::capnp::word(37, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(2, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(92, 13, 0, 0, 86, 14, 0, 0),
                ::capnp::word(21, 0, 0, 0, 154, 1, 0, 0),
                ::capnp::word(45, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 119, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                ::capnp::word(68, 101, 115, 99, 114, 105, 112, 116),
                ::capnp::word(111, 114, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 58, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(36, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(48, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(45, 0, 0, 0, 82, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(44, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(56, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(104, 97, 110, 100, 108, 101, 0, 0),
                ::capnp::word(17, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(253, 180, 16, 211, 142, 115, 184, 214),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(17, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(112, 117, 98, 108, 105, 99, 75, 101),
                ::capnp::word(121, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <crate::keystore_protocol_capnp::keystore::key::Owned as ::capnp::introspect::Introspect>::introspect(),
                    1 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[0,1];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0,1];
            pub(crate) const TYPE_ID: u64 = 0x8d6f_db63_31c9_5a8e;
        }
    }


    pub mod key {
        #![allow(unused_variables)]
        pub type IdParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::id_params::Owned>;
        pub type IdResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::id_results::Owned>;
        pub type DecryptCiphertextParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::decrypt_ciphertext_params::Owned>;
        pub type DecryptCiphertextResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::decrypt_ciphertext_results::Owned>;
        pub type SignMessageParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::sign_message_params::Owned>;
        pub type SignMessageResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::sign_message_results::Owned>;
        pub type UnlockParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::unlock_params::Owned>;
        pub type UnlockResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::unlock_results::Owned>;
        pub type AvailableParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::available_params::Owned>;
        pub type AvailableResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::available_results::Owned>;
        pub type LockedParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::locked_params::Owned>;
        pub type LockedResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::locked_results::Owned>;
        pub type DecryptionCapableParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::decryption_capable_params::Owned>;
        pub type DecryptionCapableResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::decryption_capable_results::Owned>;
        pub type SigningCapableParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::signing_capable_params::Owned>;
        pub type SigningCapableResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::signing_capable_results::Owned>;
        pub type ExportParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::export_params::Owned>;
        pub type ExportResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::export_results::Owned>;
        pub type PasswordSourceParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::password_source_params::Owned>;
        pub type PasswordSourceResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::password_source_results::Owned>;
        pub type ChangePasswordParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::change_password_params::Owned>;
        pub type ChangePasswordResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::change_password_results::Owned>;
        pub type DeleteSecretKeyMaterialParams<> = ::capnp::capability::Params<crate::keystore_protocol_capnp::keystore::key::delete_secret_key_material_params::Owned>;
        pub type DeleteSecretKeyMaterialResults<> = ::capnp::capability::Results<crate::keystore_protocol_capnp::keystore::key::delete_secret_key_material_results::Owned>;

        pub struct Client {
            pub client: ::capnp::capability::Client,
        }
        impl  ::capnp::capability::FromClientHook for Client {
            fn new(hook: Box<::capnp::capability::DynClientHook>) -> Self {
                Self { client: ::capnp::capability::Client::new(hook),  }
            }
            fn into_client_hook(self) -> Box<::capnp::capability::DynClientHook> {
                self.client.hook
            }
            fn as_client_hook(&self) -> &::capnp::capability::DynClientHook {
                &*self.client.hook
            }
        }
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Capability.into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Client; type Builder<'a> = Client; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Client; }
        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Client<>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, _default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(reader.get_capability()?))
            }
        }
        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Client<>  {
            fn init_pointer(_builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                unimplemented!()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(::capnp::capability::FromClientHook::new(builder.get_capability()?))
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Client<>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, from: Self, _canonicalize: bool) -> ::capnp::Result<()> {
                pointer.set_capability(from.client.hook);
                ::core::result::Result::Ok(())
            }
        }
        impl  ::capnp::traits::HasTypeId for Client {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl  Clone for Client {
            fn clone(&self) -> Self {
                Self { client: self.client.clone(),  }
            }
        }
        impl  Client {
            pub fn id_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::id_params::Owned,crate::keystore_protocol_capnp::keystore::key::id_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 0, ::core::option::Option::None)
            }
            pub fn decrypt_ciphertext_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::decrypt_ciphertext_params::Owned,crate::keystore_protocol_capnp::keystore::key::decrypt_ciphertext_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 1, ::core::option::Option::None)
            }
            pub fn sign_message_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::sign_message_params::Owned,crate::keystore_protocol_capnp::keystore::key::sign_message_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 2, ::core::option::Option::None)
            }
            pub fn unlock_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::unlock_params::Owned,crate::keystore_protocol_capnp::keystore::key::unlock_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 3, ::core::option::Option::None)
            }
            pub fn available_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::available_params::Owned,crate::keystore_protocol_capnp::keystore::key::available_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 4, ::core::option::Option::None)
            }
            pub fn locked_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::locked_params::Owned,crate::keystore_protocol_capnp::keystore::key::locked_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 5, ::core::option::Option::None)
            }
            pub fn decryption_capable_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::decryption_capable_params::Owned,crate::keystore_protocol_capnp::keystore::key::decryption_capable_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 6, ::core::option::Option::None)
            }
            pub fn signing_capable_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::signing_capable_params::Owned,crate::keystore_protocol_capnp::keystore::key::signing_capable_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 7, ::core::option::Option::None)
            }
            pub fn export_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::export_params::Owned,crate::keystore_protocol_capnp::keystore::key::export_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 8, ::core::option::Option::None)
            }
            pub fn password_source_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::password_source_params::Owned,crate::keystore_protocol_capnp::keystore::key::password_source_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 9, ::core::option::Option::None)
            }
            pub fn change_password_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::change_password_params::Owned,crate::keystore_protocol_capnp::keystore::key::change_password_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 10, ::core::option::Option::None)
            }
            pub fn delete_secret_key_material_request(&self) -> ::capnp::capability::Request<crate::keystore_protocol_capnp::keystore::key::delete_secret_key_material_params::Owned,crate::keystore_protocol_capnp::keystore::key::delete_secret_key_material_results::Owned> {
                self.client.new_call(_private::TYPE_ID, 11, ::core::option::Option::None)
            }
        }
        pub trait Server<> : 'static  {
            fn id(self: ::capnp::capability::Rc<Self>, _: IdParams<>, _: IdResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::id not implemented".to_string()))) }
            fn decrypt_ciphertext(self: ::capnp::capability::Rc<Self>, _: DecryptCiphertextParams<>, _: DecryptCiphertextResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::decrypt_ciphertext not implemented".to_string()))) }
            fn sign_message(self: ::capnp::capability::Rc<Self>, _: SignMessageParams<>, _: SignMessageResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::sign_message not implemented".to_string()))) }
            fn unlock(self: ::capnp::capability::Rc<Self>, _: UnlockParams<>, _: UnlockResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::unlock not implemented".to_string()))) }
            fn available(self: ::capnp::capability::Rc<Self>, _: AvailableParams<>, _: AvailableResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::available not implemented".to_string()))) }
            fn locked(self: ::capnp::capability::Rc<Self>, _: LockedParams<>, _: LockedResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::locked not implemented".to_string()))) }
            fn decryption_capable(self: ::capnp::capability::Rc<Self>, _: DecryptionCapableParams<>, _: DecryptionCapableResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::decryption_capable not implemented".to_string()))) }
            fn signing_capable(self: ::capnp::capability::Rc<Self>, _: SigningCapableParams<>, _: SigningCapableResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::signing_capable not implemented".to_string()))) }
            fn export(self: ::capnp::capability::Rc<Self>, _: ExportParams<>, _: ExportResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::export not implemented".to_string()))) }
            fn password_source(self: ::capnp::capability::Rc<Self>, _: PasswordSourceParams<>, _: PasswordSourceResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::password_source not implemented".to_string()))) }
            fn change_password(self: ::capnp::capability::Rc<Self>, _: ChangePasswordParams<>, _: ChangePasswordResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::change_password not implemented".to_string()))) }
            fn delete_secret_key_material(self: ::capnp::capability::Rc<Self>, _: DeleteSecretKeyMaterialParams<>, _: DeleteSecretKeyMaterialResults<>) -> impl ::core::future::Future<Output = Result<(), ::capnp::Error>> + 'static { ::core::future::ready(Err(::capnp::Error::unimplemented("method key::Server::delete_secret_key_material not implemented".to_string()))) }
        }
        pub struct ServerDispatch<_T,> {
            pub server: ::capnp::capability::Rc<_T>,
        }
        impl <_S: Server + 'static, > ::capnp::capability::FromServer<_S> for Client   {
            type Dispatch = ServerDispatch<_S, >;
            fn from_server(s: ::capnp::capability::Rc<_S>) -> ServerDispatch<_S, > {
                ServerDispatch { server: s,  }
            }
        }
        impl <_T: Server> ::core::ops::Deref for ServerDispatch<_T> {
            type Target = _T;
            fn deref(&self) -> &_T { &self.server}
        }
        impl <_T: Server> ::core::clone::Clone for ServerDispatch<_T> {
            fn clone(&self) -> Self { Self { server: self.server.clone(),  } }
        }
        impl <_T: Server> ::capnp::capability::Server for ServerDispatch<_T> {
            fn dispatch_call(self, interface_id: u64, method_id: u16, params: ::capnp::capability::Params<::capnp::any_pointer::Owned>, results: ::capnp::capability::Results<::capnp::any_pointer::Owned>) -> ::capnp::capability::DispatchCallResult {
                match interface_id {
                    _private::TYPE_ID => Self::dispatch_call_internal(self.server, method_id, params, results),
                    _ => { ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::err(::capnp::Error::unimplemented("Method not implemented.".to_string())), false) }
                }
            }
            fn as_ptr(&self) -> usize { ::capnp::capability::Rc::as_ptr(&self.server) as usize }
        }
        impl <_T :Server> ServerDispatch<_T> {
            pub fn dispatch_call_internal(this: ::capnp::capability::Rc<_T>, method_id: u16, params: ::capnp::capability::Params<::capnp::any_pointer::Owned>, results: ::capnp::capability::Results<::capnp::any_pointer::Owned>) -> ::capnp::capability::DispatchCallResult {
                match method_id {
                    0 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::id(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    1 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::decrypt_ciphertext(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    2 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::sign_message(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    3 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::unlock(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    4 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::available(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    5 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::locked(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    6 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::decryption_capable(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    7 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::signing_capable(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    8 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::export(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    9 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::password_source(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    10 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::change_password(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    11 => ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::from_future(<_T as Server>::delete_secret_key_material(this, ::capnp::private::capability::internal_get_typed_params(params), ::capnp::private::capability::internal_get_typed_results(results))), false),
                    _ => { ::capnp::capability::DispatchCallResult::new(::capnp::capability::Promise::err(::capnp::Error::unimplemented("Method not implemented.".to_string())), false) }
                }
            }
        }
        pub(crate) mod _private {
            pub(crate) const TYPE_ID: u64 = 0xd6b8_738e_d310_b4fd;
        }

        pub mod id_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 20] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(239, 87, 111, 32, 5, 183, 241, 233),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 154, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 105, 100, 36, 80, 97, 114, 97),
                    ::capnp::word(109, 115, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0xe9f1_b705_206f_57ef;
            }
        }

        pub mod id_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,::capnp::text::Owned>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::text::Owned>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,::capnp::text::Owned>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::text::Owned> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<::capnp::text::Owned> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 48] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(29, 46, 73, 15, 38, 16, 17, 212),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 162, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 105, 100, 36, 82, 101, 115, 117),
                    ::capnp::word(108, 116, 115, 0, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(68, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<::capnp::text::Owned> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xd411_1026_0f49_2e1d;
            }
        }

        pub mod decrypt_ciphertext_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_algo(self) -> u8 {
                    self.reader.get_data_field::<u8>(0)
                }
                #[inline]
                pub fn get_ciphertext(self) -> ::capnp::Result<::capnp::data::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_ciphertext(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
                #[inline]
                pub fn get_plaintext_len(self) -> u32 {
                    self.reader.get_data_field::<u32>(1)
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_algo(self) -> u8 {
                    self.builder.get_data_field::<u8>(0)
                }
                #[inline]
                pub fn set_algo(&mut self, value: u8)  {
                    self.builder.set_data_field::<u8>(0, value);
                }
                #[inline]
                pub fn get_ciphertext(self) -> ::capnp::Result<::capnp::data::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_ciphertext(&mut self, value: ::capnp::data::Reader<'_>)  {
                    self.builder.reborrow().get_pointer_field(0).set_data(value);
                }
                #[inline]
                pub fn init_ciphertext(self, size: u32) -> ::capnp::data::Builder<'a> {
                    self.builder.get_pointer_field(0).init_data(size)
                }
                #[inline]
                pub fn has_ciphertext(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
                #[inline]
                pub fn get_plaintext_len(self) -> u32 {
                    self.builder.get_data_field::<u32>(1)
                }
                #[inline]
                pub fn set_plaintext_len(&mut self, value: u32)  {
                    self.builder.set_data_field::<u32>(1, value);
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 70] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(19, 57, 209, 230, 148, 96, 214, 166),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 18, 2, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(45, 0, 0, 0, 175, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 100, 101, 99, 114, 121, 112, 116),
                    ::capnp::word(67, 105, 112, 104, 101, 114, 116, 101),
                    ::capnp::word(120, 116, 36, 80, 97, 114, 97, 109),
                    ::capnp::word(115, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(12, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(69, 0, 0, 0, 42, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(64, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(76, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(73, 0, 0, 0, 90, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(72, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(84, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(2, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 2, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(81, 0, 0, 0, 106, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(80, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(92, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(97, 108, 103, 111, 0, 0, 0, 0),
                    ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(99, 105, 112, 104, 101, 114, 116, 101),
                    ::capnp::word(120, 116, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(112, 108, 97, 105, 110, 116, 101, 120),
                    ::capnp::word(116, 76, 101, 110, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <u8 as ::capnp::introspect::Introspect>::introspect(),
                        1 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                        2 => <u32 as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0,1,2];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0,1,2];
                pub(crate) const TYPE_ID: u64 = 0xa6d6_6094_e6d1_3913;
            }
        }

        pub mod decrypt_ciphertext_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,::capnp::data::Owned>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::data::Owned>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,::capnp::data::Owned>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::data::Owned> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<::capnp::data::Owned> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 50] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(215, 0, 81, 82, 119, 253, 109, 253),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 26, 2, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(45, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 100, 101, 99, 114, 121, 112, 116),
                    ::capnp::word(67, 105, 112, 104, 101, 114, 116, 101),
                    ::capnp::word(120, 116, 36, 82, 101, 115, 117, 108),
                    ::capnp::word(116, 115, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(68, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<::capnp::data::Owned> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xfd6d_fd77_5251_00d7;
            }
        }

        pub mod sign_message_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_hash_algo(self) -> u8 {
                    self.reader.get_data_field::<u8>(0)
                }
                #[inline]
                pub fn get_digest(self) -> ::capnp::Result<::capnp::data::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_digest(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_hash_algo(self) -> u8 {
                    self.builder.get_data_field::<u8>(0)
                }
                #[inline]
                pub fn set_hash_algo(&mut self, value: u8)  {
                    self.builder.set_data_field::<u8>(0, value);
                }
                #[inline]
                pub fn get_digest(self) -> ::capnp::Result<::capnp::data::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_digest(&mut self, value: ::capnp::data::Reader<'_>)  {
                    self.builder.reborrow().get_pointer_field(0).set_data(value);
                }
                #[inline]
                pub fn init_digest(self, size: u32) -> ::capnp::data::Builder<'a> {
                    self.builder.get_pointer_field(0).init_data(size)
                }
                #[inline]
                pub fn has_digest(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 53] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(240, 152, 9, 246, 146, 254, 57, 251),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 226, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 119, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 115, 105, 103, 110, 77, 101, 115),
                    ::capnp::word(115, 97, 103, 101, 36, 80, 97, 114),
                    ::capnp::word(97, 109, 115, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 82, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(40, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(52, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(49, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(44, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(56, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(104, 97, 115, 104, 95, 97, 108, 103),
                    ::capnp::word(111, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(100, 105, 103, 101, 115, 116, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <u8 as ::capnp::introspect::Introspect>::introspect(),
                        1 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0,1];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[1,0];
                pub(crate) const TYPE_ID: u64 = 0xfb39_fe92_f609_98f0;
            }
        }

        pub mod sign_message_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,crate::keystore_protocol_capnp::keystore::signed_data::Owned>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,crate::keystore_protocol_capnp::keystore::signed_data::Owned>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,crate::keystore_protocol_capnp::keystore::signed_data::Owned>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,crate::keystore_protocol_capnp::keystore::signed_data::Owned> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<crate::keystore_protocol_capnp::keystore::signed_data::Owned> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 49] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(204, 212, 50, 17, 106, 140, 153, 147),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 234, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 115, 105, 103, 110, 77, 101, 115),
                    ::capnp::word(115, 97, 103, 101, 36, 82, 101, 115),
                    ::capnp::word(117, 108, 116, 115, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(68, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(176, 251, 46, 224, 17, 72, 132, 154),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<crate::keystore_protocol_capnp::keystore::signed_data::Owned> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0x9399_8c6a_1132_d4cc;
            }
        }

        pub mod unlock_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_password(self) -> ::capnp::Result<::capnp::data::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_password(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_password(self) -> ::capnp::Result<::capnp::data::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_password(&mut self, value: ::capnp::data::Reader<'_>)  {
                    self.builder.reborrow().get_pointer_field(0).set_data(value);
                }
                #[inline]
                pub fn init_password(self, size: u32) -> ::capnp::data::Builder<'a> {
                    self.builder.get_pointer_field(0).init_data(size)
                }
                #[inline]
                pub fn has_password(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 37] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(57, 32, 196, 69, 96, 136, 87, 247),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 186, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 117, 110, 108, 111, 99, 107, 36),
                    ::capnp::word(80, 97, 114, 97, 109, 115, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 74, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(12, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(24, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(112, 97, 115, 115, 119, 111, 114, 100),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xf757_8860_45c4_2039;
            }
        }

        pub mod unlock_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::void_result::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::void_result::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::void_result::Reader<'_>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::void_result::Builder<'a> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::void_result::Pipeline {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 36] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(104, 33, 215, 199, 71, 247, 162, 244),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 194, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 117, 110, 108, 111, 99, 107, 36),
                    ::capnp::word(82, 101, 115, 117, 108, 116, 115, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(20, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(177, 193, 154, 58, 150, 13, 229, 220),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::void_result::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xf4a2_f747_c7d7_2168;
            }
        }

        pub mod available_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 21] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(32, 89, 35, 74, 42, 159, 251, 151),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 210, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 97, 118, 97, 105, 108, 97, 98),
                    ::capnp::word(108, 101, 36, 80, 97, 114, 97, 109),
                    ::capnp::word(115, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0x97fb_9f2a_4a23_5920;
            }
        }

        pub mod available_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::bool_result::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::bool_result::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::bool_result::Reader<'_>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::bool_result::Builder<'a> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::bool_result::Pipeline {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 37] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(131, 17, 140, 168, 33, 110, 133, 244),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 218, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 97, 118, 97, 105, 108, 97, 98),
                    ::capnp::word(108, 101, 36, 82, 101, 115, 117, 108),
                    ::capnp::word(116, 115, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(20, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(2, 163, 131, 229, 159, 50, 103, 128),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::bool_result::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xf485_6e21_a88c_1183;
            }
        }

        pub mod locked_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 20] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(1, 232, 25, 174, 196, 91, 142, 224),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 186, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 108, 111, 99, 107, 101, 100, 36),
                    ::capnp::word(80, 97, 114, 97, 109, 115, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0xe08e_5bc4_ae19_e801;
            }
        }

        pub mod locked_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,crate::keystore_protocol_capnp::keystore::protection::Owned>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,crate::keystore_protocol_capnp::keystore::protection::Owned>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,crate::keystore_protocol_capnp::keystore::protection::Owned>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,crate::keystore_protocol_capnp::keystore::protection::Owned> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<crate::keystore_protocol_capnp::keystore::protection::Owned> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 48] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(155, 93, 130, 167, 73, 235, 240, 211),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 194, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 108, 111, 99, 107, 101, 100, 36),
                    ::capnp::word(82, 101, 115, 117, 108, 116, 115, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(68, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(22, 64, 230, 113, 197, 117, 207, 194),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<crate::keystore_protocol_capnp::keystore::protection::Owned> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xd3f0_eb49_a782_5d9b;
            }
        }

        pub mod decryption_capable_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 22] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(117, 243, 22, 44, 117, 152, 240, 136),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 18, 2, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 100, 101, 99, 114, 121, 112, 116),
                    ::capnp::word(105, 111, 110, 67, 97, 112, 97, 98),
                    ::capnp::word(108, 101, 36, 80, 97, 114, 97, 109),
                    ::capnp::word(115, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0x88f0_9875_2c16_f375;
            }
        }

        pub mod decryption_capable_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::bool_result::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::bool_result::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::bool_result::Reader<'_>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::bool_result::Builder<'a> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::bool_result::Pipeline {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 38] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(49, 193, 69, 161, 170, 176, 0, 192),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 26, 2, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(45, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 100, 101, 99, 114, 121, 112, 116),
                    ::capnp::word(105, 111, 110, 67, 97, 112, 97, 98),
                    ::capnp::word(108, 101, 36, 82, 101, 115, 117, 108),
                    ::capnp::word(116, 115, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(20, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(2, 163, 131, 229, 159, 50, 103, 128),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::bool_result::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xc000_b0aa_a145_c131;
            }
        }

        pub mod signing_capable_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 21] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(233, 202, 79, 192, 123, 106, 150, 223),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 250, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 115, 105, 103, 110, 105, 110, 103),
                    ::capnp::word(67, 97, 112, 97, 98, 108, 101, 36),
                    ::capnp::word(80, 97, 114, 97, 109, 115, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0xdf96_6a7b_c04f_cae9;
            }
        }

        pub mod signing_capable_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::bool_result::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::bool_result::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::bool_result::Reader<'_>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::bool_result::Builder<'a> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::bool_result::Pipeline {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 37] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(12, 235, 22, 9, 137, 155, 90, 235),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 2, 2, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 115, 105, 103, 110, 105, 110, 103),
                    ::capnp::word(67, 97, 112, 97, 98, 108, 101, 36),
                    ::capnp::word(82, 101, 115, 117, 108, 116, 115, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(20, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(2, 163, 131, 229, 159, 50, 103, 128),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::bool_result::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xeb5a_9b89_0916_eb0c;
            }
        }

        pub mod export_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 20] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(71, 24, 179, 221, 248, 183, 209, 162),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 186, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 101, 120, 112, 111, 114, 116, 36),
                    ::capnp::word(80, 97, 114, 97, 109, 115, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0xa2d1_b7f8_ddb3_1847;
            }
        }

        pub mod export_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,::capnp::data::Owned>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::data::Owned>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,::capnp::data::Owned>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::data::Owned> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<::capnp::data::Owned> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 48] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(11, 188, 108, 12, 85, 115, 159, 243),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 194, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 101, 120, 112, 111, 114, 116, 36),
                    ::capnp::word(82, 101, 115, 117, 108, 116, 115, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(68, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<::capnp::data::Owned> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xf39f_7355_0c6c_bc0b;
            }
        }

        pub mod password_source_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 21] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(137, 150, 203, 76, 182, 165, 93, 226),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 250, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 112, 97, 115, 115, 119, 111, 114),
                    ::capnp::word(100, 83, 111, 117, 114, 99, 101, 36),
                    ::capnp::word(80, 97, 114, 97, 109, 115, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0xe25d_a5b6_4ccb_9689;
            }
        }

        pub mod password_source_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,crate::keystore_protocol_capnp::keystore::password_source::Owned>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,crate::keystore_protocol_capnp::keystore::password_source::Owned>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,crate::keystore_protocol_capnp::keystore::password_source::Owned>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,crate::keystore_protocol_capnp::keystore::password_source::Owned> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<crate::keystore_protocol_capnp::keystore::password_source::Owned> {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 49] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(215, 77, 217, 143, 43, 65, 59, 163),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 2, 2, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 112, 97, 115, 115, 119, 111, 114),
                    ::capnp::word(100, 83, 111, 117, 114, 99, 101, 36),
                    ::capnp::word(82, 101, 115, 117, 108, 116, 115, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(68, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                    ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(105, 154, 107, 247, 78, 28, 227, 149),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::result::Owned<crate::keystore_protocol_capnp::keystore::password_source::Owned> as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xa33b_412b_8fd9_4dd7;
            }
        }

        pub mod change_password_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_password(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::change_password_action::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_password(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_password(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::change_password_action::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_password(&mut self, value: crate::keystore_protocol_capnp::keystore::change_password_action::Reader<'_>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_password(self, ) -> crate::keystore_protocol_capnp::keystore::change_password_action::Builder<'a> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_password(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_password(&self) -> crate::keystore_protocol_capnp::keystore::change_password_action::Pipeline {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 38] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(241, 151, 29, 42, 121, 68, 105, 204),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 250, 1, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 99, 104, 97, 110, 103, 101, 80),
                    ::capnp::word(97, 115, 115, 119, 111, 114, 100, 36),
                    ::capnp::word(80, 97, 114, 97, 109, 115, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 74, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(12, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(24, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(112, 97, 115, 115, 119, 111, 114, 100),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(111, 90, 122, 181, 181, 54, 115, 159),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::change_password_action::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xcc69_4479_2a1d_97f1;
            }
        }

        pub mod change_password_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::void_result::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::void_result::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::void_result::Reader<'_>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::void_result::Builder<'a> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::void_result::Pipeline {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 37] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(101, 245, 14, 19, 75, 38, 131, 161),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 2, 2, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(41, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 99, 104, 97, 110, 103, 101, 80),
                    ::capnp::word(97, 115, 115, 119, 111, 114, 100, 36),
                    ::capnp::word(82, 101, 115, 117, 108, 116, 115, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(20, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(177, 193, 154, 58, 150, 13, 229, 220),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::void_result::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xa183_264b_130e_f565;
            }
        }

        pub mod delete_secret_key_material_params {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <> Reader<'_,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 22] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(242, 101, 207, 202, 53, 129, 221, 181),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 66, 2, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 100, 101, 108, 101, 116, 101, 83),
                    ::capnp::word(101, 99, 114, 101, 116, 75, 101, 121),
                    ::capnp::word(77, 97, 116, 101, 114, 105, 97, 108),
                    ::capnp::word(36, 80, 97, 114, 97, 109, 115, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_field_index(index)
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
                pub(crate) const TYPE_ID: u64 = 0xb5dd_8135_cacf_65f2;
            }
        }

        pub mod delete_secret_key_material_results {
            #[derive(Copy, Clone)]
            pub struct Owned(());
            impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
            impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
            impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

            pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
            impl <> ::core::marker::Copy for Reader<'_,>  {}
            impl <> ::core::clone::Clone for Reader<'_,>  {
                fn clone(&self) -> Self { *self }
            }

            impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
                fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                    Self { reader,  }
                }
            }

            impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
                fn from(reader: Reader<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <> ::core::fmt::Debug for Reader<'_,>  {
                fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                    core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
                }
            }

            impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
                fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(reader.get_struct(default)?.into())
                }
            }

            impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
                fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                    self.reader
                }
            }

            impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
                fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                    self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
                }
            }

            impl <'a,> Reader<'a,>  {
                pub fn reborrow(&self) -> Reader<'_,> {
                    Self { .. *self }
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.reader.total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::void_result::Reader<'a>> {
                    ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.reader.get_pointer_field(0).is_null()
                }
            }

            pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
            impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
                const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
            }
            impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
                const TYPE_ID: u64 = _private::TYPE_ID;
            }
            impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
                fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                    Self { builder,  }
                }
            }

            impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
                fn from(builder: Builder<'a,>) -> Self {
                    Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
                }
            }

            impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
                fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                    self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
                }
            }

            impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
                fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                    builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
                }
                fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                    ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
                }
            }

            impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
                fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
            }

            impl <'a,> Builder<'a,>  {
                pub fn into_reader(self) -> Reader<'a,> {
                    self.builder.into_reader().into()
                }
                pub fn reborrow(&mut self) -> Builder<'_,> {
                    Builder { builder: self.builder.reborrow() }
                }
                pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                    self.builder.as_reader().into()
                }

                pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                    self.builder.as_reader().total_size()
                }
                #[inline]
                pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::void_result::Builder<'a>> {
                    ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                }
                #[inline]
                pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::void_result::Reader<'_>) -> ::capnp::Result<()> {
                    ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
                }
                #[inline]
                pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::void_result::Builder<'a> {
                    ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
                }
                #[inline]
                pub fn has_result(&self) -> bool {
                    !self.builder.is_pointer_field_null(0)
                }
            }

            pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
            impl ::capnp::capability::FromTypelessPipeline for Pipeline {
                fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                    Self { _typeless: typeless,  }
                }
            }
            impl Pipeline  {
                pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::void_result::Pipeline {
                    ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
                }
            }
            mod _private {
                pub(crate) static ENCODED_NODE: [::capnp::Word; 39] = [
                    ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                    ::capnp::word(191, 216, 220, 15, 147, 199, 215, 168),
                    ::capnp::word(41, 0, 0, 0, 1, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(21, 0, 0, 0, 74, 2, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(49, 0, 0, 0, 63, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                    ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                    ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                    ::capnp::word(116, 111, 114, 101, 46, 75, 101, 121),
                    ::capnp::word(46, 100, 101, 108, 101, 116, 101, 83),
                    ::capnp::word(101, 99, 114, 101, 116, 75, 101, 121),
                    ::capnp::word(77, 97, 116, 101, 114, 105, 97, 108),
                    ::capnp::word(36, 82, 101, 115, 117, 108, 116, 115),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                    ::capnp::word(20, 0, 0, 0, 2, 0, 1, 0),
                    ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(177, 193, 154, 58, 150, 13, 229, 220),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                    ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ];
                pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                    match index {
                        0 => <crate::keystore_protocol_capnp::keystore::void_result::Owned as ::capnp::introspect::Introspect>::introspect(),
                        _ => ::capnp::introspect::panic_invalid_field_index(index),
                    }
                }
                pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                    ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
                }
                pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
                pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                    &ARENA,
                    NONUNION_MEMBERS,
                    MEMBERS_BY_DISCRIMINANT,
                    MEMBERS_BY_NAME
                );
                pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
                pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
                pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
                pub(crate) const TYPE_ID: u64 = 0xa8d7_c793_0fdc_d8bf;
            }
        }
    }

    pub mod protection {
        pub use self::Which::{Unlocked,UnknownProtection,Password,ExternalPassword,ExternalTouch,ExternalOther};

        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn has_unknown_protection(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 1 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_password(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 2 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_external_password(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 3 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_external_touch(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 4 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_external_other(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 5 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichReader<'a,>, ::capnp::NotInSchema> {
                match self.reader.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Unlocked(
                            ()
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(UnknownProtection(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    2 => {
                        ::core::result::Result::Ok(Password(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    3 => {
                        ::core::result::Result::Ok(ExternalPassword(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    4 => {
                        ::core::result::Result::Ok(ExternalTouch(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    5 => {
                        ::core::result::Result::Ok(ExternalOther(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 1 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn set_unlocked(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 0);
            }
            #[inline]
            pub fn set_unknown_protection(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 1);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_unknown_protection(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 1);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_unknown_protection(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 1 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_password(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 2);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_password(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 2);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_password(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 2 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_external_password(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 3);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_external_password(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 3);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_external_password(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 3 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_external_touch(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 4);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_external_touch(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 4);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_external_touch(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 4 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_external_other(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 5);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_external_other(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 5);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_external_other(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 5 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichBuilder<'a,>, ::capnp::NotInSchema> {
                match self.builder.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Unlocked(
                            ()
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(UnknownProtection(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    2 => {
                        ::core::result::Result::Ok(Password(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    3 => {
                        ::core::result::Result::Ok(ExternalPassword(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    4 => {
                        ::core::result::Result::Ok(ExternalTouch(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    5 => {
                        ::core::result::Result::Ok(ExternalOther(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 119] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(22, 64, 230, 113, 197, 117, 207, 194),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(1, 0, 7, 0, 0, 0, 6, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(187, 22, 0, 0, 71, 30, 0, 0),
                ::capnp::word(21, 0, 0, 0, 130, 1, 0, 0),
                ::capnp::word(41, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 0, 0, 0, 87, 1, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 80, 114, 111),
                ::capnp::word(116, 101, 99, 116, 105, 111, 110, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(24, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 255, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(153, 0, 0, 0, 74, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(152, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(164, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 254, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(161, 0, 0, 0, 146, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(164, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(176, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(2, 0, 253, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 2, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(173, 0, 0, 0, 74, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(172, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(184, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(3, 0, 252, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 3, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(181, 0, 0, 0, 138, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(184, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(196, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(4, 0, 251, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 4, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(193, 0, 0, 0, 114, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(192, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(204, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(5, 0, 250, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 5, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(201, 0, 0, 0, 114, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(200, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(212, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(117, 110, 108, 111, 99, 107, 101, 100),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(117, 110, 107, 110, 111, 119, 110, 80),
                ::capnp::word(114, 111, 116, 101, 99, 116, 105, 111),
                ::capnp::word(110, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(112, 97, 115, 115, 119, 111, 114, 100),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 120, 116, 101, 114, 110, 97, 108),
                ::capnp::word(80, 97, 115, 115, 119, 111, 114, 100),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 120, 116, 101, 114, 110, 97, 108),
                ::capnp::word(84, 111, 117, 99, 104, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 120, 116, 101, 114, 110, 97, 108),
                ::capnp::word(79, 116, 104, 101, 114, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <() as ::capnp::introspect::Introspect>::introspect(),
                    1 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    2 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    3 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    4 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    5 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[0,1,2,3,4,5];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[5,3,4,2,1,0];
            pub(crate) const TYPE_ID: u64 = 0xc2cf_75c5_71e6_4016;
        }
        pub enum Which<A0,A1,A2,A3,A4> {
            Unlocked(()),
            UnknownProtection(A0),
            Password(A1),
            ExternalPassword(A2),
            ExternalTouch(A3),
            ExternalOther(A4),
        }
        pub type WhichReader<'a,> = Which<::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>>;
        pub type WhichBuilder<'a,> = Which<::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>>;
    }

    pub mod password_source {
        pub use self::Which::{Inline,InlineWithConfirmation,ExternalOnDemand,ExternalSideEffect};

        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <> Reader<'_,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichReader, ::capnp::NotInSchema> {
                match self.reader.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Inline(
                            ()
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(InlineWithConfirmation(
                            ()
                        ))
                    }
                    2 => {
                        ::core::result::Result::Ok(ExternalOnDemand(
                            ()
                        ))
                    }
                    3 => {
                        ::core::result::Result::Ok(ExternalSideEffect(
                            ()
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 0 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn set_inline(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 0);
            }
            #[inline]
            pub fn set_inline_with_confirmation(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 1);
            }
            #[inline]
            pub fn set_external_on_demand(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 2);
            }
            #[inline]
            pub fn set_external_side_effect(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 3);
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichBuilder, ::capnp::NotInSchema> {
                match self.builder.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Inline(
                            ()
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(InlineWithConfirmation(
                            ()
                        ))
                    }
                    2 => {
                        ::core::result::Result::Ok(ExternalOnDemand(
                            ()
                        ))
                    }
                    3 => {
                        ::core::result::Result::Ok(ExternalSideEffect(
                            ()
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 88] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(105, 154, 107, 247, 78, 28, 227, 149),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(0, 0, 7, 0, 0, 0, 4, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(75, 30, 0, 0, 5, 38, 0, 0),
                ::capnp::word(21, 0, 0, 0, 162, 1, 0, 0),
                ::capnp::word(45, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 231, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 80, 97, 115),
                ::capnp::word(115, 119, 111, 114, 100, 83, 111, 117),
                ::capnp::word(114, 99, 101, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(16, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 255, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(97, 0, 0, 0, 58, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(92, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(104, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 254, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 0, 0, 0, 186, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(104, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(116, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(2, 0, 253, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 2, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(113, 0, 0, 0, 138, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(116, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(128, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(3, 0, 252, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 3, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(125, 0, 0, 0, 154, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(128, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(140, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(105, 110, 108, 105, 110, 101, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(105, 110, 108, 105, 110, 101, 87, 105),
                ::capnp::word(116, 104, 67, 111, 110, 102, 105, 114),
                ::capnp::word(109, 97, 116, 105, 111, 110, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 120, 116, 101, 114, 110, 97, 108),
                ::capnp::word(79, 110, 68, 101, 109, 97, 110, 100),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 120, 116, 101, 114, 110, 97, 108),
                ::capnp::word(83, 105, 100, 101, 69, 102, 102, 101),
                ::capnp::word(99, 116, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <() as ::capnp::introspect::Introspect>::introspect(),
                    1 => <() as ::capnp::introspect::Introspect>::introspect(),
                    2 => <() as ::capnp::introspect::Introspect>::introspect(),
                    3 => <() as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[0,1,2,3];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[2,3,0,1];
            pub(crate) const TYPE_ID: u64 = 0x95e3_1c4e_f76b_9a69;
        }
        pub enum Which {
            Inline(()),
            InlineWithConfirmation(()),
            ExternalOnDemand(()),
            ExternalSideEffect(()),
        }
        pub type WhichReader = Which;
        pub type WhichBuilder = Which;
    }

    pub mod change_password_action {
        pub use self::Which::{Password,Ask};

        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn has_password(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 0 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichReader<'a,>, ::capnp::NotInSchema> {
                match self.reader.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Password(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Ask(
                            ()
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 1 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn set_password(&mut self, value: ::capnp::data::Reader<'_>)  {
                self.builder.set_data_field::<u16>(0, 0);
                self.builder.reborrow().get_pointer_field(0).set_data(value);
            }
            #[inline]
            pub fn init_password(self, size: u32) -> ::capnp::data::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 0);
                self.builder.get_pointer_field(0).init_data(size)
            }
            #[inline]
            pub fn has_password(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 0 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_ask(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 1);
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichBuilder<'a,>, ::capnp::NotInSchema> {
                match self.builder.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Password(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Ask(
                            ()
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 54] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(111, 90, 122, 181, 181, 54, 115, 159),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(1, 0, 7, 0, 0, 0, 2, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(9, 38, 0, 0, 177, 39, 0, 0),
                ::capnp::word(21, 0, 0, 0, 210, 1, 0, 0),
                ::capnp::word(49, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(45, 0, 0, 0, 119, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 67, 104, 97),
                ::capnp::word(110, 103, 101, 80, 97, 115, 115, 119),
                ::capnp::word(111, 114, 100, 65, 99, 116, 105, 111),
                ::capnp::word(110, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 255, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 74, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(40, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(52, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 254, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(49, 0, 0, 0, 34, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(44, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(56, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(112, 97, 115, 115, 119, 111, 114, 100),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(97, 115, 107, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                    1 => <() as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[0,1];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[1,0];
            pub(crate) const TYPE_ID: u64 = 0x9f73_36b5_b57a_5a6f;
        }
        pub enum Which<A0> {
            Password(A0),
            Ask(()),
        }
        pub type WhichReader<'a,> = Which<::capnp::Result<::capnp::data::Reader<'a>>>;
        pub type WhichBuilder<'a,> = Which<::capnp::Result<::capnp::data::Builder<'a>>>;
    }

    pub mod signed_data {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn get_pk_algo(self) -> u8 {
                self.reader.get_data_field::<u8>(0)
            }
            #[inline]
            pub fn get_mpis(self) -> ::capnp::Result<::capnp::data::Reader<'a>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_mpis(&self) -> bool {
                !self.reader.get_pointer_field(0).is_null()
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 1 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn get_pk_algo(self) -> u8 {
                self.builder.get_data_field::<u8>(0)
            }
            #[inline]
            pub fn set_pk_algo(&mut self, value: u8)  {
                self.builder.set_data_field::<u8>(0, value);
            }
            #[inline]
            pub fn get_mpis(self) -> ::capnp::Result<::capnp::data::Builder<'a>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_mpis(&mut self, value: ::capnp::data::Reader<'_>)  {
                self.builder.reborrow().get_pointer_field(0).set_data(value);
            }
            #[inline]
            pub fn init_mpis(self, size: u32) -> ::capnp::data::Builder<'a> {
                self.builder.get_pointer_field(0).init_data(size)
            }
            #[inline]
            pub fn has_mpis(&self) -> bool {
                !self.builder.is_pointer_field_null(0)
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 51] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(176, 251, 46, 224, 17, 72, 132, 154),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(181, 39, 0, 0, 54, 40, 0, 0),
                ::capnp::word(21, 0, 0, 0, 130, 1, 0, 0),
                ::capnp::word(41, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 0, 0, 0, 119, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 83, 105, 103),
                ::capnp::word(110, 101, 100, 68, 97, 116, 97, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 58, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(36, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(48, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(45, 0, 0, 0, 42, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(40, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(52, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(112, 107, 65, 108, 103, 111, 0, 0),
                ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(6, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(109, 112, 105, 115, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <u8 as ::capnp::introspect::Introspect>::introspect(),
                    1 => <::capnp::data::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[0,1];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[1,0];
            pub(crate) const TYPE_ID: u64 = 0x9a84_4811_e02e_fbb0;
        }
    }

    pub mod unit {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <> Reader<'_,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 20] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(194, 174, 151, 64, 118, 168, 55, 248),
                ::capnp::word(37, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(58, 40, 0, 0, 111, 40, 0, 0),
                ::capnp::word(21, 0, 0, 0, 82, 1, 0, 0),
                ::capnp::word(41, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 85, 110, 105),
                ::capnp::word(116, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_field_index(index)
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
            pub(crate) const TYPE_ID: u64 = 0xf837_a876_4097_aec2;
        }
    }

    pub mod error {
        pub use self::Which::{GenericError,Protocol,Eof,InaccessibleDecryptionKey,NotDecryptionCapable,NotSigningCapable,InternalError,ExternalImportRequired,SecretKeyMaterialSealed,NoInlinePassword,NoExternalPassword};

        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn has_generic_error(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 0 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_inaccessible_decryption_key(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 3 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_not_decryption_capable(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 4 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_not_signing_capable(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 5 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_internal_error(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 6 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_external_import_required(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 7 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_secret_key_material_sealed(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 8 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_no_inline_password(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 9 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_no_external_password(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 10 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichReader<'a,>, ::capnp::NotInSchema> {
                match self.reader.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(GenericError(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Protocol(
                            ()
                        ))
                    }
                    2 => {
                        ::core::result::Result::Ok(Eof(
                            ()
                        ))
                    }
                    3 => {
                        ::core::result::Result::Ok(InaccessibleDecryptionKey(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    4 => {
                        ::core::result::Result::Ok(NotDecryptionCapable(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    5 => {
                        ::core::result::Result::Ok(NotSigningCapable(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    6 => {
                        ::core::result::Result::Ok(InternalError(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    7 => {
                        ::core::result::Result::Ok(ExternalImportRequired(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    8 => {
                        ::core::result::Result::Ok(SecretKeyMaterialSealed(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    9 => {
                        ::core::result::Result::Ok(NoInlinePassword(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    10 => {
                        ::core::result::Result::Ok(NoExternalPassword(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 1 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn set_generic_error(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 0);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_generic_error(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 0);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_generic_error(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 0 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_protocol(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 1);
            }
            #[inline]
            pub fn set_eof(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 2);
            }
            #[inline]
            pub fn set_inaccessible_decryption_key(&mut self, value: ::capnp::struct_list::Reader<'_,crate::keystore_protocol_capnp::keystore::inaccessible_decryption_key::Owned>) -> ::capnp::Result<()> {
                self.builder.set_data_field::<u16>(0, 3);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_inaccessible_decryption_key(self, size: u32) -> ::capnp::struct_list::Builder<'a,crate::keystore_protocol_capnp::keystore::inaccessible_decryption_key::Owned> {
                self.builder.set_data_field::<u16>(0, 3);
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), size)
            }
            #[inline]
            pub fn has_inaccessible_decryption_key(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 3 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_not_decryption_capable(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 4);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_not_decryption_capable(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 4);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_not_decryption_capable(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 4 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_not_signing_capable(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 5);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_not_signing_capable(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 5);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_not_signing_capable(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 5 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_internal_error(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 6);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_internal_error(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 6);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_internal_error(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 6 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_external_import_required(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 7);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_external_import_required(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 7);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_external_import_required(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 7 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_secret_key_material_sealed(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 8);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_secret_key_material_sealed(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 8);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_secret_key_material_sealed(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 8 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_no_inline_password(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 9);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_no_inline_password(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 9);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_no_inline_password(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 9 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_no_external_password(&mut self, value: impl ::capnp::traits::SetterInput<::capnp::text::Owned>)  {
                self.builder.set_data_field::<u16>(0, 10);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false).unwrap()
            }
            #[inline]
            pub fn init_no_external_password(self, size: u32) -> ::capnp::text::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 10);
                self.builder.get_pointer_field(0).init_text(size)
            }
            #[inline]
            pub fn has_no_external_password(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 10 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichBuilder<'a,>, ::capnp::NotInSchema> {
                match self.builder.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(GenericError(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Protocol(
                            ()
                        ))
                    }
                    2 => {
                        ::core::result::Result::Ok(Eof(
                            ()
                        ))
                    }
                    3 => {
                        ::core::result::Result::Ok(InaccessibleDecryptionKey(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    4 => {
                        ::core::result::Result::Ok(NotDecryptionCapable(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    5 => {
                        ::core::result::Result::Ok(NotSigningCapable(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    6 => {
                        ::core::result::Result::Ok(InternalError(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    7 => {
                        ::core::result::Result::Ok(ExternalImportRequired(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    8 => {
                        ::core::result::Result::Ok(SecretKeyMaterialSealed(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    9 => {
                        ::core::result::Result::Ok(NoInlinePassword(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    10 => {
                        ::core::result::Result::Ok(NoExternalPassword(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 208] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(209, 122, 213, 61, 87, 194, 223, 241),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(1, 0, 7, 0, 0, 0, 11, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(114, 40, 0, 0, 1, 50, 0, 0),
                ::capnp::word(21, 0, 0, 0, 90, 1, 0, 0),
                ::capnp::word(41, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 0, 0, 0, 111, 2, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 69, 114, 114),
                ::capnp::word(111, 114, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(44, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 255, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 1, 0, 0, 106, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(36, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(48, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 254, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(45, 1, 0, 0, 74, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(44, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(56, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(2, 0, 253, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 2, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(53, 1, 0, 0, 34, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(48, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(60, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(3, 0, 252, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 3, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(57, 1, 0, 0, 210, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(64, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(92, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(4, 0, 251, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 4, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(89, 1, 0, 0, 170, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(92, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(104, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(5, 0, 250, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 5, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 1, 0, 0, 146, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(104, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(116, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(6, 0, 249, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 6, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(113, 1, 0, 0, 114, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(112, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(124, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(7, 0, 248, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(121, 1, 0, 0, 186, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(124, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(136, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(8, 0, 247, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 8, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(133, 1, 0, 0, 194, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(136, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(148, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(9, 0, 246, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 9, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(145, 1, 0, 0, 138, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(148, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(160, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(10, 0, 245, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 10, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(157, 1, 0, 0, 154, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(160, 1, 0, 0, 3, 0, 1, 0),
                ::capnp::word(172, 1, 0, 0, 2, 0, 1, 0),
                ::capnp::word(103, 101, 110, 101, 114, 105, 99, 69),
                ::capnp::word(114, 114, 111, 114, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(112, 114, 111, 116, 111, 99, 111, 108),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 111, 102, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(105, 110, 97, 99, 99, 101, 115, 115),
                ::capnp::word(105, 98, 108, 101, 68, 101, 99, 114),
                ::capnp::word(121, 112, 116, 105, 111, 110, 75, 101),
                ::capnp::word(121, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(14, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(183, 143, 143, 82, 232, 118, 133, 161),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(14, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(110, 111, 116, 68, 101, 99, 114, 121),
                ::capnp::word(112, 116, 105, 111, 110, 67, 97, 112),
                ::capnp::word(97, 98, 108, 101, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(110, 111, 116, 83, 105, 103, 110, 105),
                ::capnp::word(110, 103, 67, 97, 112, 97, 98, 108),
                ::capnp::word(101, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(105, 110, 116, 101, 114, 110, 97, 108),
                ::capnp::word(69, 114, 114, 111, 114, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 120, 116, 101, 114, 110, 97, 108),
                ::capnp::word(73, 109, 112, 111, 114, 116, 82, 101),
                ::capnp::word(113, 117, 105, 114, 101, 100, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 101, 99, 114, 101, 116, 75, 101),
                ::capnp::word(121, 77, 97, 116, 101, 114, 105, 97),
                ::capnp::word(108, 83, 101, 97, 108, 101, 100, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(110, 111, 73, 110, 108, 105, 110, 101),
                ::capnp::word(80, 97, 115, 115, 119, 111, 114, 100),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(110, 111, 69, 120, 116, 101, 114, 110),
                ::capnp::word(97, 108, 80, 97, 115, 115, 119, 111),
                ::capnp::word(114, 100, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(12, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    1 => <() as ::capnp::introspect::Introspect>::introspect(),
                    2 => <() as ::capnp::introspect::Introspect>::introspect(),
                    3 => <::capnp::struct_list::Owned<crate::keystore_protocol_capnp::keystore::inaccessible_decryption_key::Owned> as ::capnp::introspect::Introspect>::introspect(),
                    4 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    5 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    6 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    7 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    8 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    9 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    10 => <::capnp::text::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[0,1,2,3,4,5,6,7,8,9,10];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[2,7,0,3,6,10,9,4,5,1,8];
            pub(crate) const TYPE_ID: u64 = 0xf1df_c257_3dd5_7ad1;
        }
        pub enum Which<A0,A1,A2,A3,A4,A5,A6,A7,A8> {
            GenericError(A0),
            Protocol(()),
            Eof(()),
            InaccessibleDecryptionKey(A1),
            NotDecryptionCapable(A2),
            NotSigningCapable(A3),
            InternalError(A4),
            ExternalImportRequired(A5),
            SecretKeyMaterialSealed(A6),
            NoInlinePassword(A7),
            NoExternalPassword(A8),
        }
        pub type WhichReader<'a,> = Which<::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::struct_list::Reader<'a,crate::keystore_protocol_capnp::keystore::inaccessible_decryption_key::Owned>>,::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>,::capnp::Result<::capnp::text::Reader<'a>>>;
        pub type WhichBuilder<'a,> = Which<::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::struct_list::Builder<'a,crate::keystore_protocol_capnp::keystore::inaccessible_decryption_key::Owned>>,::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>,::capnp::Result<::capnp::text::Builder<'a>>>;
    }

    pub mod result { /* T */
        pub use self::Which::{Ok,Err};

        #[derive(Copy, Clone)]
        pub struct Owned<T> {
            _phantom: ::core::marker::PhantomData<T>
        }
        impl <T> ::capnp::introspect::Introspect for Owned <T> where T: ::capnp::traits::Owned  { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<T>, annotation_types: _private::get_annotation_types::<T> }).into() } }
        impl <T> ::capnp::traits::Owned for Owned <T> where T: ::capnp::traits::Owned  { type Reader<'a> = Reader<'a, T>; type Builder<'a> = Builder<'a, T>; }
        impl <T> ::capnp::traits::OwnedStruct for Owned <T> where T: ::capnp::traits::Owned  { type Reader<'a> = Reader<'a, T>; type Builder<'a> = Builder<'a, T>; }
        impl <T> ::capnp::traits::Pipelined for Owned<T> where T: ::capnp::traits::Owned  { type Pipeline = Pipeline<T>; }

        pub struct Reader<'a,T> where T: ::capnp::traits::Owned  {
            reader: ::capnp::private::layout::StructReader<'a>,
            _phantom: ::core::marker::PhantomData<T>
        }
        impl <T> ::core::marker::Copy for Reader<'_,T> where T: ::capnp::traits::Owned  {}
        impl <T> ::core::clone::Clone for Reader<'_,T> where T: ::capnp::traits::Owned  {
            fn clone(&self) -> Self { *self }
        }

        impl <T> ::capnp::traits::HasTypeId for Reader<'_,T> where T: ::capnp::traits::Owned  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,T> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,T> where T: ::capnp::traits::Owned  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader, _phantom: ::core::marker::PhantomData, }
            }
        }

        impl <'a,T> ::core::convert::From<Reader<'a,T>> for ::capnp::dynamic_value::Reader<'a> where T: ::capnp::traits::Owned  {
            fn from(reader: Reader<'a,T>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<T>, annotation_types: _private::get_annotation_types::<T>})))
            }
        }

        impl <T> ::core::fmt::Debug for Reader<'_,T> where T: ::capnp::traits::Owned  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,T> ::capnp::traits::FromPointerReader<'a> for Reader<'a,T> where T: ::capnp::traits::Owned  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,T> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,T> where T: ::capnp::traits::Owned  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,T> ::capnp::traits::Imbue<'a> for Reader<'a,T> where T: ::capnp::traits::Owned  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,T> Reader<'a,T> where T: ::capnp::traits::Owned  {
            pub fn reborrow(&self) -> Reader<'_,T> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn has_ok(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 0 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn has_err(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 1 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichReader<'a,T>, ::capnp::NotInSchema> {
                match self.reader.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Ok(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Err(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Builder<'a,T> where T: ::capnp::traits::Owned  {
            builder: ::capnp::private::layout::StructBuilder<'a>,
            _phantom: ::core::marker::PhantomData<T>
        }
        impl <T> ::capnp::traits::HasStructSize for Builder<'_,T> where T: ::capnp::traits::Owned  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 1 };
        }
        impl <T> ::capnp::traits::HasTypeId for Builder<'_,T> where T: ::capnp::traits::Owned  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,T> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,T> where T: ::capnp::traits::Owned  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder, _phantom: ::core::marker::PhantomData, }
            }
        }

        impl <'a,T> ::core::convert::From<Builder<'a,T>> for ::capnp::dynamic_value::Builder<'a> where T: ::capnp::traits::Owned  {
            fn from(builder: Builder<'a,T>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<T>, annotation_types: _private::get_annotation_types::<T>})))
            }
        }

        impl <'a,T> ::capnp::traits::ImbueMut<'a> for Builder<'a,T> where T: ::capnp::traits::Owned  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,T> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,T> where T: ::capnp::traits::Owned  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <T> ::capnp::traits::SetterInput<Owned<T>> for Reader<'_,T> where T: ::capnp::traits::Owned  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,T> Builder<'a,T> where T: ::capnp::traits::Owned  {
            pub fn into_reader(self) -> Reader<'a,T> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,T> {
                Builder { builder: self.builder.reborrow(), ..*self }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,T> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn initn_ok(self, length: u32) -> <T as ::capnp::traits::Owned>::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 0);
                ::capnp::any_pointer::Builder::new(self.builder.get_pointer_field(0)).initn_as(length)
            }
            #[inline]
            pub fn set_ok(&mut self, value: impl ::capnp::traits::SetterInput<T>) -> ::capnp::Result<()> {
                self.builder.set_data_field::<u16>(0, 0);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_ok(self, ) -> <T as ::capnp::traits::Owned>::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 0);
                ::capnp::any_pointer::Builder::new(self.builder.get_pointer_field(0)).init_as()
            }
            #[inline]
            pub fn has_ok(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 0 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn set_err(&mut self, value: crate::keystore_protocol_capnp::keystore::error::Reader<'_>) -> ::capnp::Result<()> {
                self.builder.set_data_field::<u16>(0, 1);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_err(self, ) -> crate::keystore_protocol_capnp::keystore::error::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 1);
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
            }
            #[inline]
            pub fn has_err(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 1 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichBuilder<'a,T>, ::capnp::NotInSchema> {
                match self.builder.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Ok(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Err(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Pipeline<T> {
            _typeless: ::capnp::any_pointer::Pipeline,
            _phantom: ::core::marker::PhantomData<T>
        }
        impl<T> ::capnp::capability::FromTypelessPipeline for Pipeline<T> {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless, _phantom: ::core::marker::PhantomData, }
            }
        }
        impl<T> Pipeline<T> where T: ::capnp::traits::Pipelined, <T as ::capnp::traits::Pipelined>::Pipeline: ::capnp::capability::FromTypelessPipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 54] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(1, 0, 7, 0, 0, 0, 2, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(5, 50, 0, 0, 82, 50, 0, 0),
                ::capnp::word(21, 0, 0, 0, 98, 1, 0, 0),
                ::capnp::word(41, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 0, 0, 0, 119, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(153, 0, 0, 0, 15, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 82, 101, 115),
                ::capnp::word(117, 108, 116, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 255, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 26, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(36, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(48, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 254, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(45, 0, 0, 0, 34, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(40, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(52, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(111, 107, 0, 0, 0, 0, 0, 0),
                ::capnp::word(18, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(18, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 114, 114, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(209, 122, 213, 61, 87, 194, 223, 241),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(4, 0, 0, 0, 0, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 18, 0, 0, 0),
                ::capnp::word(84, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types<T>(index: u16) -> ::capnp::introspect::Type where T: ::capnp::traits::Owned  {
                match index {
                    0 => <T as ::capnp::introspect::Introspect>::introspect(),
                    1 => <crate::keystore_protocol_capnp::keystore::error::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types<T>(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type where T: ::capnp::traits::Owned  {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[0,1];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[1,0];
            pub(crate) const TYPE_ID: u64 = 0xdecc_42a4_f7cc_4d58;
        }
        pub enum Which<A0,A1> {
            Ok(A0),
            Err(A1),
        }
        pub type WhichReader<'a,T> = Which<::capnp::Result<<T as ::capnp::traits::Owned>::Reader<'a>>,::capnp::Result<crate::keystore_protocol_capnp::keystore::error::Reader<'a>>>;
        pub type WhichBuilder<'a,T> = Which<::capnp::Result<<T as ::capnp::traits::Owned>::Builder<'a>>,::capnp::Result<crate::keystore_protocol_capnp::keystore::error::Builder<'a>>>;
    }

    pub mod void_result {
        pub use self::Which::{Ok,Err};

        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn has_err(&self) -> bool {
                if self.reader.get_data_field::<u16>(0) != 1 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichReader<'a,>, ::capnp::NotInSchema> {
                match self.reader.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Ok(
                            ()
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Err(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 1 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn set_ok(&mut self, _value: ())  {
                self.builder.set_data_field::<u16>(0, 0);
            }
            #[inline]
            pub fn set_err(&mut self, value: crate::keystore_protocol_capnp::keystore::error::Reader<'_>) -> ::capnp::Result<()> {
                self.builder.set_data_field::<u16>(0, 1);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_err(self, ) -> crate::keystore_protocol_capnp::keystore::error::Builder<'a> {
                self.builder.set_data_field::<u16>(0, 1);
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
            }
            #[inline]
            pub fn has_err(&self) -> bool {
                if self.builder.get_data_field::<u16>(0) != 1 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichBuilder<'a,>, ::capnp::NotInSchema> {
                match self.builder.get_data_field::<u16>(0) {
                    0 => {
                        ::core::result::Result::Ok(Ok(
                            ()
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Err(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 51] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(177, 193, 154, 58, 150, 13, 229, 220),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(1, 0, 7, 0, 0, 0, 2, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(86, 50, 0, 0, 23, 52, 0, 0),
                ::capnp::word(21, 0, 0, 0, 130, 1, 0, 0),
                ::capnp::word(41, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 0, 0, 0, 119, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 86, 111, 105),
                ::capnp::word(100, 82, 101, 115, 117, 108, 116, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 255, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 26, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(36, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(48, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 254, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(45, 0, 0, 0, 34, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(40, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(52, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(111, 107, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 114, 114, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(209, 122, 213, 61, 87, 194, 223, 241),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <() as ::capnp::introspect::Introspect>::introspect(),
                    1 => <crate::keystore_protocol_capnp::keystore::error::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[0,1];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[1,0];
            pub(crate) const TYPE_ID: u64 = 0xdce5_0d96_3a9a_c1b1;
        }
        pub enum Which<A0> {
            Ok(()),
            Err(A0),
        }
        pub type WhichReader<'a,> = Which<::capnp::Result<crate::keystore_protocol_capnp::keystore::error::Reader<'a>>>;
        pub type WhichBuilder<'a,> = Which<::capnp::Result<crate::keystore_protocol_capnp::keystore::error::Builder<'a>>>;
    }

    pub mod bool_result {
        pub use self::Which::{Ok,Err};

        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn has_err(&self) -> bool {
                if self.reader.get_data_field::<u16>(1) != 1 { return false; }
                !self.reader.get_pointer_field(0).is_null()
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichReader<'a,>, ::capnp::NotInSchema> {
                match self.reader.get_data_field::<u16>(1) {
                    0 => {
                        ::core::result::Result::Ok(Ok(
                            self.reader.get_bool_field(0)
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Err(
                            ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 1, pointers: 1 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn set_ok(&mut self, value: bool)  {
                self.builder.set_data_field::<u16>(1, 0);
                self.builder.set_bool_field(0, value);
            }
            #[inline]
            pub fn set_err(&mut self, value: crate::keystore_protocol_capnp::keystore::error::Reader<'_>) -> ::capnp::Result<()> {
                self.builder.set_data_field::<u16>(1, 1);
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_err(self, ) -> crate::keystore_protocol_capnp::keystore::error::Builder<'a> {
                self.builder.set_data_field::<u16>(1, 1);
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
            }
            #[inline]
            pub fn has_err(&self) -> bool {
                if self.builder.get_data_field::<u16>(1) != 1 { return false; }
                !self.builder.is_pointer_field_null(0)
            }
            #[inline]
            pub fn which(self) -> ::core::result::Result<WhichBuilder<'a,>, ::capnp::NotInSchema> {
                match self.builder.get_data_field::<u16>(1) {
                    0 => {
                        ::core::result::Result::Ok(Ok(
                            self.builder.get_bool_field(0)
                        ))
                    }
                    1 => {
                        ::core::result::Result::Ok(Err(
                            ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
                        ))
                    }
                    x => ::core::result::Result::Err(::capnp::NotInSchema(x))
                }
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 51] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(2, 163, 131, 229, 159, 50, 103, 128),
                ::capnp::word(37, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(175, 101, 136, 71, 15, 117, 240, 215),
                ::capnp::word(1, 0, 7, 0, 0, 0, 2, 0),
                ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(27, 52, 0, 0, 222, 53, 0, 0),
                ::capnp::word(21, 0, 0, 0, 130, 1, 0, 0),
                ::capnp::word(41, 0, 0, 0, 7, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 0, 0, 0, 119, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 66, 111, 111),
                ::capnp::word(108, 82, 101, 115, 117, 108, 116, 0),
                ::capnp::word(0, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 255, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(41, 0, 0, 0, 26, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(36, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(48, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(1, 0, 254, 255, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(45, 0, 0, 0, 34, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(40, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(52, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(111, 107, 0, 0, 0, 0, 0, 0),
                ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(101, 114, 114, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(209, 122, 213, 61, 87, 194, 223, 241),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <bool as ::capnp::introspect::Introspect>::introspect(),
                    1 => <crate::keystore_protocol_capnp::keystore::error::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[0,1];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[1,0];
            pub(crate) const TYPE_ID: u64 = 0x8067_329f_e583_a302;
        }
        pub enum Which<A0> {
            Ok(bool),
            Err(A0),
        }
        pub type WhichReader<'a,> = Which<::capnp::Result<crate::keystore_protocol_capnp::keystore::error::Reader<'a>>>;
        pub type WhichBuilder<'a,> = Which<::capnp::Result<crate::keystore_protocol_capnp::keystore::error::Builder<'a>>>;
    }

    pub mod backends_params {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <> Reader<'_,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 0 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 20] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(81, 148, 193, 56, 168, 123, 133, 244),
                ::capnp::word(37, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(21, 0, 0, 0, 170, 1, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 98, 97, 99),
                ::capnp::word(107, 101, 110, 100, 115, 36, 80, 97),
                ::capnp::word(114, 97, 109, 115, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_field_index(index)
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[];
            pub(crate) const TYPE_ID: u64 = 0xf485_7ba8_38c1_9451;
        }
    }

    pub mod backends_results {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::backend::Client>>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_result(&self) -> bool {
                !self.reader.get_pointer_field(0).is_null()
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::backend::Client>>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::backend::Client>>) -> ::capnp::Result<()> {
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::backend::Client>> {
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
            }
            #[inline]
            pub fn has_result(&self) -> bool {
                !self.builder.is_pointer_field_null(0)
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
            pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::backend::Client>> {
                ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
            }
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 52] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(196, 7, 227, 7, 59, 241, 123, 172),
                ::capnp::word(37, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(21, 0, 0, 0, 178, 1, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 98, 97, 99),
                ::capnp::word(107, 101, 110, 100, 115, 36, 82, 101),
                ::capnp::word(115, 117, 108, 116, 115, 0, 0, 0),
                ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(84, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(14, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(17, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(16, 233, 206, 234, 45, 65, 200, 139),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <crate::keystore_protocol_capnp::keystore::result::Owned<::capnp::capability_list::Owned<crate::keystore_protocol_capnp::keystore::backend::Client>> as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
            pub(crate) const TYPE_ID: u64 = 0xac7b_f13b_07e3_07c4;
        }
    }

    pub mod decrypt_params {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn get_pkesks(self) -> ::capnp::Result<::capnp::data_list::Reader<'a>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_pkesks(&self) -> bool {
                !self.reader.get_pointer_field(0).is_null()
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn get_pkesks(self) -> ::capnp::Result<::capnp::data_list::Builder<'a>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_pkesks(&mut self, value: ::capnp::data_list::Reader<'_>) -> ::capnp::Result<()> {
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_pkesks(self, size: u32) -> ::capnp::data_list::Builder<'a> {
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), size)
            }
            #[inline]
            pub fn has_pkesks(&self) -> bool {
                !self.builder.is_pointer_field_null(0)
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 40] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(222, 162, 216, 174, 218, 0, 222, 140),
                ::capnp::word(37, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(21, 0, 0, 0, 162, 1, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 100, 101, 99),
                ::capnp::word(114, 121, 112, 116, 36, 80, 97, 114),
                ::capnp::word(97, 109, 115, 0, 0, 0, 0, 0),
                ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(36, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(112, 107, 101, 115, 107, 115, 0, 0),
                ::capnp::word(14, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(13, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(14, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <::capnp::data_list::Owned as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
            pub(crate) const TYPE_ID: u64 = 0x8cde_00da_aed8_a2de;
        }
    }

    pub mod decrypt_results {
        #[derive(Copy, Clone)]
        pub struct Owned(());
        impl ::capnp::introspect::Introspect for Owned { fn introspect() -> ::capnp::introspect::Type { ::capnp::introspect::TypeVariant::Struct(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types, annotation_types: _private::get_annotation_types }).into() } }
        impl ::capnp::traits::Owned for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::OwnedStruct for Owned { type Reader<'a> = Reader<'a>; type Builder<'a> = Builder<'a>; }
        impl ::capnp::traits::Pipelined for Owned { type Pipeline = Pipeline; }

        pub struct Reader<'a> { reader: ::capnp::private::layout::StructReader<'a> }
        impl <> ::core::marker::Copy for Reader<'_,>  {}
        impl <> ::core::clone::Clone for Reader<'_,>  {
            fn clone(&self) -> Self { *self }
        }

        impl <> ::capnp::traits::HasTypeId for Reader<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructReader<'a>> for Reader<'a,>  {
            fn from(reader: ::capnp::private::layout::StructReader<'a>) -> Self {
                Self { reader,  }
            }
        }

        impl <'a,> ::core::convert::From<Reader<'a,>> for ::capnp::dynamic_value::Reader<'a>  {
            fn from(reader: Reader<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Reader::new(reader.reader, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <> ::core::fmt::Debug for Reader<'_,>  {
            fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::result::Result<(), ::core::fmt::Error> {
                core::fmt::Debug::fmt(&::core::convert::Into::<::capnp::dynamic_value::Reader<'_>>::into(*self), f)
            }
        }

        impl <'a,> ::capnp::traits::FromPointerReader<'a> for Reader<'a,>  {
            fn get_from_pointer(reader: &::capnp::private::layout::PointerReader<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(reader.get_struct(default)?.into())
            }
        }

        impl <'a,> ::capnp::traits::IntoInternalStructReader<'a> for Reader<'a,>  {
            fn into_internal_struct_reader(self) -> ::capnp::private::layout::StructReader<'a> {
                self.reader
            }
        }

        impl <'a,> ::capnp::traits::Imbue<'a> for Reader<'a,>  {
            fn imbue(&mut self, cap_table: &'a ::capnp::private::layout::CapTable) {
                self.reader.imbue(::capnp::private::layout::CapTableReader::Plain(cap_table))
            }
        }

        impl <'a,> Reader<'a,>  {
            pub fn reborrow(&self) -> Reader<'_,> {
                Self { .. *self }
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.reader.total_size()
            }
            #[inline]
            pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Reader<'a,crate::keystore_protocol_capnp::keystore::decrypted_pkesk::Owned>> {
                ::capnp::traits::FromPointerReader::get_from_pointer(&self.reader.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn has_result(&self) -> bool {
                !self.reader.get_pointer_field(0).is_null()
            }
        }

        pub struct Builder<'a> { builder: ::capnp::private::layout::StructBuilder<'a> }
        impl <> ::capnp::traits::HasStructSize for Builder<'_,>  {
            const STRUCT_SIZE: ::capnp::private::layout::StructSize = ::capnp::private::layout::StructSize { data: 0, pointers: 1 };
        }
        impl <> ::capnp::traits::HasTypeId for Builder<'_,>  {
            const TYPE_ID: u64 = _private::TYPE_ID;
        }
        impl <'a,> ::core::convert::From<::capnp::private::layout::StructBuilder<'a>> for Builder<'a,>  {
            fn from(builder: ::capnp::private::layout::StructBuilder<'a>) -> Self {
                Self { builder,  }
            }
        }

        impl <'a,> ::core::convert::From<Builder<'a,>> for ::capnp::dynamic_value::Builder<'a>  {
            fn from(builder: Builder<'a,>) -> Self {
                Self::Struct(::capnp::dynamic_struct::Builder::new(builder.builder, ::capnp::schema::StructSchema::new(::capnp::introspect::RawBrandedStructSchema { generic: &_private::RAW_SCHEMA, field_types: _private::get_field_types::<>, annotation_types: _private::get_annotation_types::<>})))
            }
        }

        impl <'a,> ::capnp::traits::ImbueMut<'a> for Builder<'a,>  {
            fn imbue_mut(&mut self, cap_table: &'a mut ::capnp::private::layout::CapTable) {
                self.builder.imbue(::capnp::private::layout::CapTableBuilder::Plain(cap_table))
            }
        }

        impl <'a,> ::capnp::traits::FromPointerBuilder<'a> for Builder<'a,>  {
            fn init_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, _size: u32) -> Self {
                builder.init_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE).into()
            }
            fn get_from_pointer(builder: ::capnp::private::layout::PointerBuilder<'a>, default: ::core::option::Option<&'a [::capnp::Word]>) -> ::capnp::Result<Self> {
                ::core::result::Result::Ok(builder.get_struct(<Self as ::capnp::traits::HasStructSize>::STRUCT_SIZE, default)?.into())
            }
        }

        impl <> ::capnp::traits::SetterInput<Owned<>> for Reader<'_,>  {
            fn set_pointer_builder(mut pointer: ::capnp::private::layout::PointerBuilder<'_>, value: Self, canonicalize: bool) -> ::capnp::Result<()> { pointer.set_struct(&value.reader, canonicalize) }
        }

        impl <'a,> Builder<'a,>  {
            pub fn into_reader(self) -> Reader<'a,> {
                self.builder.into_reader().into()
            }
            pub fn reborrow(&mut self) -> Builder<'_,> {
                Builder { builder: self.builder.reborrow() }
            }
            pub fn reborrow_as_reader(&self) -> Reader<'_,> {
                self.builder.as_reader().into()
            }

            pub fn total_size(&self) -> ::capnp::Result<::capnp::MessageSize> {
                self.builder.as_reader().total_size()
            }
            #[inline]
            pub fn get_result(self) -> ::capnp::Result<crate::keystore_protocol_capnp::keystore::result::Builder<'a,crate::keystore_protocol_capnp::keystore::decrypted_pkesk::Owned>> {
                ::capnp::traits::FromPointerBuilder::get_from_pointer(self.builder.get_pointer_field(0), ::core::option::Option::None)
            }
            #[inline]
            pub fn set_result(&mut self, value: crate::keystore_protocol_capnp::keystore::result::Reader<'_,crate::keystore_protocol_capnp::keystore::decrypted_pkesk::Owned>) -> ::capnp::Result<()> {
                ::capnp::traits::SetterInput::set_pointer_builder(self.builder.reborrow().get_pointer_field(0), value, false)
            }
            #[inline]
            pub fn init_result(self, ) -> crate::keystore_protocol_capnp::keystore::result::Builder<'a,crate::keystore_protocol_capnp::keystore::decrypted_pkesk::Owned> {
                ::capnp::traits::FromPointerBuilder::init_pointer(self.builder.get_pointer_field(0), 0)
            }
            #[inline]
            pub fn has_result(&self) -> bool {
                !self.builder.is_pointer_field_null(0)
            }
        }

        pub struct Pipeline { _typeless: ::capnp::any_pointer::Pipeline }
        impl ::capnp::capability::FromTypelessPipeline for Pipeline {
            fn new(typeless: ::capnp::any_pointer::Pipeline) -> Self {
                Self { _typeless: typeless,  }
            }
        }
        impl Pipeline  {
            pub fn get_result(&self) -> crate::keystore_protocol_capnp::keystore::result::Pipeline<crate::keystore_protocol_capnp::keystore::decrypted_pkesk::Owned> {
                ::capnp::capability::FromTypelessPipeline::new(self._typeless.get_pointer_field(0))
            }
        }
        mod _private {
            pub(crate) static ENCODED_NODE: [::capnp::Word; 48] = [
                ::capnp::word(0, 0, 0, 0, 6, 0, 6, 0),
                ::capnp::word(216, 27, 135, 20, 215, 80, 75, 252),
                ::capnp::word(37, 0, 0, 0, 1, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(1, 0, 7, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(21, 0, 0, 0, 170, 1, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(37, 0, 0, 0, 63, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(115, 114, 99, 47, 107, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 95, 112, 114, 111),
                ::capnp::word(116, 111, 99, 111, 108, 46, 99, 97),
                ::capnp::word(112, 110, 112, 58, 75, 101, 121, 115),
                ::capnp::word(116, 111, 114, 101, 46, 100, 101, 99),
                ::capnp::word(114, 121, 112, 116, 36, 82, 101, 115),
                ::capnp::word(117, 108, 116, 115, 0, 0, 0, 0),
                ::capnp::word(4, 0, 0, 0, 3, 0, 4, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 1, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(13, 0, 0, 0, 58, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(8, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(68, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(114, 101, 115, 117, 108, 116, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 31, 0, 0, 0),
                ::capnp::word(4, 0, 0, 0, 2, 0, 1, 0),
                ::capnp::word(88, 77, 204, 247, 164, 66, 204, 222),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(1, 0, 0, 0, 23, 0, 0, 0),
                ::capnp::word(4, 0, 0, 0, 1, 0, 1, 0),
                ::capnp::word(1, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 3, 0, 1, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(135, 169, 134, 148, 114, 56, 38, 210),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(16, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
                ::capnp::word(0, 0, 0, 0, 0, 0, 0, 0),
            ];
            pub(crate) fn get_field_types(index: u16) -> ::capnp::introspect::Type {
                match index {
                    0 => <crate::keystore_protocol_capnp::keystore::result::Owned<crate::keystore_protocol_capnp::keystore::decrypted_pkesk::Owned> as ::capnp::introspect::Introspect>::introspect(),
                    _ => ::capnp::introspect::panic_invalid_field_index(index),
                }
            }
            pub(crate) fn get_annotation_types(child_index: Option<u16>, index: u32) -> ::capnp::introspect::Type {
                ::capnp::introspect::panic_invalid_annotation_indices(child_index, index)
            }
            pub(crate) static ARENA: ::capnp::private::arena::GeneratedCodeArena = ::capnp::private::arena::GeneratedCodeArena::new(&ENCODED_NODE);
            pub(crate) static RAW_SCHEMA: ::capnp::introspect::RawStructSchema = ::capnp::introspect::RawStructSchema::new(
                &ARENA,
                NONUNION_MEMBERS,
                MEMBERS_BY_DISCRIMINANT,
                MEMBERS_BY_NAME
            );
            pub(crate) static NONUNION_MEMBERS : &[u16] = &[0];
            pub(crate) static MEMBERS_BY_DISCRIMINANT : &[u16] = &[];
            pub(crate) static MEMBERS_BY_NAME : &[u16] = &[0];
            pub(crate) const TYPE_ID: u64 = 0xfc4b_50d7_1487_1bd8;
        }
    }
}
